#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from botorch.utils.testing import BotorchTestCase
from botorch_community.posteriors.riemann import (
    BoundedRiemannPosterior,
    MultivariateRiemannPosterior,
)


class TestRiemannPosterior(BotorchTestCase):
    def test_properties(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            borders = torch.linspace(0, 1, 6, **tkwargs)
            probabilities = torch.tensor([0.1, 0.2, 0.1, 0.2, 0.4], **tkwargs)
            posterior = BoundedRiemannPosterior(borders, probabilities)
            self.assertTrue(torch.equal(posterior.borders, borders))
            self.assertTrue(torch.equal(posterior.probabilities, probabilities))
            self.assertEqual(posterior.dtype, dtype)
            self.assertEqual(posterior.device, self.device)

    def test_integrate(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            borders = torch.tensor([0.0, 0.5, 1.5], **tkwargs)
            probabilities = torch.tensor([0.2, 0.8], **tkwargs)
            posterior = BoundedRiemannPosterior(borders, probabilities)

            def ag_integrate(lower, upper):
                return (upper + lower) / 2

            result = posterior.integrate(ag_integrate)
            expected_result = (0.0 + 0.5) / 2 * 0.2 / 0.5 + (1.5 + 0.5) / 2 * 0.8 / 1.0
            self.assertLess((result - expected_result).abs(), 1e-5)

    def test_rsample(self):
        torch.manual_seed(13)
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}

            # Create Riemann Posterior based on a normal distribution
            dist_mean = torch.tensor(3.0, **tkwargs)
            dist_std = torch.tensor(2.0, **tkwargs)
            test_distribution = torch.distributions.Normal(dist_mean, dist_std)
            test_samples = test_distribution.rsample(torch.Size([10000]))

            n_buckets = 5000
            borders = torch.quantile(
                test_samples, torch.linspace(0, 1, n_buckets + 1, **tkwargs)
            )
            probabilities = torch.ones(n_buckets, **tkwargs) / n_buckets
            posterior = BoundedRiemannPosterior(borders, probabilities)

            # Check that the mean and variance of the samples are correct
            samples = posterior.rsample(torch.Size([5000000]))
            self.assertLess((samples.mean(dim=0) - dist_mean).abs().item(), 0.02)
            self.assertLess((samples.std(dim=0) - dist_std).abs().item(), 0.02)

            # Create Riemann Posterior based on a Gamma distribution
            dist_alpha = torch.tensor(2.0, **tkwargs)
            dist_beta = torch.tensor(4.5, **tkwargs)
            test_distribution = torch.distributions.Gamma(dist_alpha, dist_beta)
            test_samples = test_distribution.rsample(torch.Size([10000]))
            n_buckets = 1000
            borders = torch.quantile(
                test_samples, torch.linspace(0, 1, n_buckets + 1, **tkwargs)
            )
            probabilities = torch.ones(n_buckets, **tkwargs) / n_buckets
            posterior = BoundedRiemannPosterior(borders, probabilities)

            # Check that the mean and variance of the samples are correct
            samples = posterior.rsample(torch.Size([10000]))
            self.assertLess(
                (samples.mean(dim=0) - test_distribution.mean).abs().item(), 0.02
            )
            self.assertLess(
                (samples.std(dim=0).pow(2) - test_distribution.variance).abs().item(),
                0.02,
            )

    def test_rsample_from_base_samples(self):
        torch.manual_seed(13)
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}

            borders = torch.linspace(-10, 10, 26, **tkwargs)
            probabilities = torch.rand(3, 25, **tkwargs)
            probabilities = probabilities / probabilities.sum(-1, keepdim=True)
            posterior = BoundedRiemannPosterior(
                borders,
                probabilities.expand(
                    12, 3, 25
                ),  # Two batch dimensions, with shapes 12 and 3
            )

            base_samples = torch.rand(30, 12, 3)  # 30 samples
            samples = posterior.rsample_from_base_samples(
                torch.Size([30, 12, 3]), base_samples
            )
            self.assertEqual(samples.shape, torch.Size([30, 12, 3, 1]))

            with self.assertRaises(ValueError):
                posterior.rsample_from_base_samples(torch.Size([10, 4]), base_samples)

    def test_mean(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            borders = torch.linspace(0, 1, 6, **tkwargs)
            probabilities = torch.tensor([0.1, 0.2, 0.1, 0.2, 0.4], **tkwargs)
            posterior = BoundedRiemannPosterior(borders, probabilities)
            mean = posterior.mean
            expected_mean = ((borders[1:] + borders[:-1]) / 2 * probabilities).sum(-1)
            self.assertLess((mean - expected_mean).abs().sum().item(), 1e-5)

    def test_variance(self):
        torch.manual_seed(13)
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}

            # Create Riemann Posterior based on a normal distribution with non-zero mean
            dist_mean = torch.tensor(2.5, **tkwargs)
            dist_std = torch.tensor(1.5, **tkwargs)
            test_distribution = torch.distributions.Normal(dist_mean, dist_std)
            test_samples = test_distribution.rsample(torch.Size([100000]))
            n_buckets = 5000
            borders = torch.quantile(
                test_samples, torch.linspace(0, 1, n_buckets + 1, **tkwargs)
            )
            probabilities = torch.ones(n_buckets, **tkwargs) / n_buckets
            posterior = BoundedRiemannPosterior(borders, probabilities)

            # Check that the variance approximately matches the true variance
            true_variance = test_distribution.variance
            computed_variance = posterior.variance
            self.assertLess((computed_variance - true_variance).abs().item(), 0.05)

            # Test with a different distribution (non-zero mean, different variance)
            dist_mean = torch.tensor(-1.0, **tkwargs)
            dist_std = torch.tensor(0.8, **tkwargs)
            test_distribution = torch.distributions.Normal(dist_mean, dist_std)
            test_samples = test_distribution.rsample(torch.Size([100000]))
            n_buckets = 3000
            borders = torch.quantile(
                test_samples, torch.linspace(0, 1, n_buckets + 1, **tkwargs)
            )
            probabilities = torch.ones(n_buckets, **tkwargs) / n_buckets
            posterior = BoundedRiemannPosterior(borders, probabilities)

            # Check that the variance approximately matches the true variance
            true_variance = test_distribution.variance
            computed_variance = posterior.variance
            self.assertLess((computed_variance - true_variance).abs().item(), 0.05)

            # Check with batch dimension
            probabilities = torch.rand(2, n_buckets, **tkwargs)
            probabilities = probabilities / probabilities.sum(-1, keepdim=True)
            posterior = BoundedRiemannPosterior(borders, probabilities)
            self.assertEqual(posterior.variance.shape, torch.Size([2, 1]))
            self.assertEqual(posterior.mean.shape, torch.Size([2, 1]))

    def test_confidence_region(self):
        torch.manual_seed(13)
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}

            # Create Riemann Posterior based on a normal distribution
            dist_mean = torch.tensor(0.0, **tkwargs)
            dist_std = torch.tensor(1.0, **tkwargs)
            test_distribution = torch.distributions.Normal(dist_mean, dist_std)
            test_samples = test_distribution.rsample(torch.Size([100000]))
            n_buckets = 5000
            borders = torch.quantile(
                test_samples, torch.linspace(0, 1, n_buckets + 1, **tkwargs)
            )
            probabilities = torch.ones(n_buckets, **tkwargs) / n_buckets
            probabilities = probabilities.unsqueeze(0)
            posterior = BoundedRiemannPosterior(borders, probabilities)

            lower, upper = posterior.confidence_region(confidence_level=0.954)
            self.assertLess((lower - (dist_mean - 2 * dist_std)).abs().item(), 0.02)
            self.assertLess((upper - (dist_mean + 2 * dist_std)).abs().item(), 0.02)

    def test_icdf(self):
        torch.manual_seed(13)
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            probabilities = torch.tensor([0.5, 0.5], **tkwargs)
            probabilities = probabilities.unsqueeze(0).unsqueeze(0)
            posterior = BoundedRiemannPosterior(
                borders=torch.tensor([0.0, 1.0, 2.0], **tkwargs),
                probabilities=probabilities,
            )
            value = torch.tensor([0.6], **tkwargs)
            value = value.unsqueeze(0).unsqueeze(0)
            true_res = torch.tensor([1.2], **tkwargs)
            self.assertTrue(
                torch.allclose(posterior.icdf(value=value).squeeze(), true_res)
            )
            # And batch dimensions (5, 12)
            probabilities = probabilities.expand(torch.Size([5, 12, 2]))
            value = value.expand(torch.Size([30, 5, 12]))
            posterior = BoundedRiemannPosterior(
                borders=torch.tensor([0.0, 1.0, 2.0], **tkwargs),
                probabilities=probabilities,
            )
            res = posterior.icdf(value=value)
            self.assertEqual(res.shape, torch.Size([30, 5, 12, 1]))
            true_res = true_res.expand(30, 5, 12, 1)
            self.assertTrue(torch.allclose(res, true_res))
            # Test with float
            value = 0.6
            true_res = torch.tensor([1.2], **tkwargs)
            self.assertTrue(
                torch.allclose(posterior.icdf(value=value).squeeze(), true_res)
            )


class TestMultivariateRiemannPosterior(BotorchTestCase):
    def test_multivariate_rsample(self):
        borders = torch.linspace(0, 1, 101)
        probabilities = torch.ones(2, 3, 100)
        probabilities = probabilities / probabilities.sum(-1, keepdim=True)
        correlation_matrix = torch.cat(
            [
                torch.eye(3, 3).unsqueeze(0),
                torch.tensor(
                    [[1.0, 0.9, 0.0], [0.9, 1.0, 0.0], [0.0, 0.0, 1.0]]
                ).unsqueeze(0),
            ],
            dim=0,
        )
        posterior = MultivariateRiemannPosterior(
            borders=borders,
            probabilities=probabilities,
            correlation_matrix=correlation_matrix,
        )
        base_samples = torch.rand(5000, 2, 3)
        samples = posterior.rsample_from_base_samples(
            torch.Size([5000, 2, 3]), base_samples
        )
        self.assertEqual(samples.shape, torch.Size([5000, 2, 3, 1]))
        # Uncorrelated samples
        self.assertTrue(
            torch.abs(torch.corrcoef(samples[:, 0, :].squeeze().T)[0, 1]).item() < 0.2
        )
        # Correlated samples
        self.assertTrue(
            torch.abs(torch.corrcoef(samples[:, 1, :].squeeze().T)[0, 1]).item() > 0.8
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Cross-validation utilities using batch evaluation mode.
"""

from __future__ import annotations

from typing import Any, NamedTuple

import torch
from botorch.exceptions.errors import UnsupportedError
from botorch.fit import fit_gpytorch_mll
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.multitask import MultiTaskGP
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
from torch import Tensor


class CVFolds(NamedTuple):
    train_X: Tensor
    test_X: Tensor
    train_Y: Tensor
    test_Y: Tensor
    train_Yvar: Tensor | None = None
    test_Yvar: Tensor | None = None


class CVResults(NamedTuple):
    model: GPyTorchModel
    posterior: GPyTorchPosterior
    observed_Y: Tensor
    observed_Yvar: Tensor | None = None


def gen_loo_cv_folds(
    train_X: Tensor, train_Y: Tensor, train_Yvar: Tensor | None = None
) -> CVFolds:
    r"""Generate LOO CV folds w.r.t. to `n`.

    Args:
        train_X: A `n x d` or `batch_shape x n x d` (batch mode) tensor of training
            features.
        train_Y: A `n x (m)` or `batch_shape x n x (m)` (batch mode) tensor of
            training observations.
        train_Yvar: An `n x (m)` or `batch_shape x n x (m)` (batch mode) tensor
            of observed measurement noise.

    Returns:
        CVFolds NamedTuple with the following fields:

        - train_X: A `n x (n-1) x d` or `batch_shape x n x (n-1) x d` tensor of
          training features.
        - test_X: A `n x 1 x d` or `batch_shape x n x 1 x d` tensor of test features.
        - train_Y: A `n x (n-1) x m` or `batch_shape x n x (n-1) x m` tensor of
          training observations.
        - test_Y: A `n x 1 x m` or `batch_shape x n x 1 x m` tensor of test
          observations.
        - train_Yvar: A `n x (n-1) x m` or `batch_shape x n x (n-1) x m` tensor
          of observed measurement noise.
        - test_Yvar: A `n x 1 x m` or `batch_shape x n x 1 x m` tensor of observed
          measurement noise.

    Example:
        >>> train_X = torch.rand(10, 1)
        >>> train_Y = torch.rand_like(train_X)
        >>> cv_folds = gen_loo_cv_folds(train_X, train_Y)
        >>> cv_folds.train_X.shape
        torch.Size([10, 9, 1])
    """
    masks = torch.eye(train_X.shape[-2], dtype=torch.uint8, device=train_X.device)
    masks = masks.to(dtype=torch.bool)
    if train_Y.dim() < train_X.dim():
        # add output dimension
        train_Y = train_Y.unsqueeze(-1)
        if train_Yvar is not None:
            train_Yvar = train_Yvar.unsqueeze(-1)
    train_X_cv = torch.cat(
        [train_X[..., ~m, :].unsqueeze(dim=-3) for m in masks], dim=-3
    )
    test_X_cv = torch.cat([train_X[..., m, :].unsqueeze(dim=-3) for m in masks], dim=-3)
    train_Y_cv = torch.cat(
        [train_Y[..., ~m, :].unsqueeze(dim=-3) for m in masks], dim=-3
    )
    test_Y_cv = torch.cat([train_Y[..., m, :].unsqueeze(dim=-3) for m in masks], dim=-3)
    if train_Yvar is None:
        train_Yvar_cv = None
        test_Yvar_cv = None
    else:
        train_Yvar_cv = torch.cat(
            [train_Yvar[..., ~m, :].unsqueeze(dim=-3) for m in masks], dim=-3
        )
        test_Yvar_cv = torch.cat(
            [train_Yvar[..., m, :].unsqueeze(dim=-3) for m in masks], dim=-3
        )
    return CVFolds(
        train_X=train_X_cv,
        test_X=test_X_cv,
        train_Y=train_Y_cv,
        test_Y=test_Y_cv,
        train_Yvar=train_Yvar_cv,
        test_Yvar=test_Yvar_cv,
    )


def batch_cross_validation(
    model_cls: type[GPyTorchModel],
    mll_cls: type[MarginalLogLikelihood],
    cv_folds: CVFolds,
    fit_args: dict[str, Any] | None = None,
    observation_noise: bool = False,
    model_init_kwargs: dict[str, Any] | None = None,
) -> CVResults:
    r"""Perform cross validation by using GPyTorch batch mode.

    WARNING: This function is currently very memory inefficient; use it only
        for problems of small size.

    Args:
        model_cls: A GPyTorchModel class. This class must initialize the likelihood
            internally. Note: Multi-task GPs are not currently supported.
        mll_cls: A MarginalLogLikelihood class.
        cv_folds: A CVFolds tuple.
        fit_args: Arguments passed along to fit_gpytorch_mll.
        model_init_kwargs: Keyword arguments passed to the model constructor.

    Returns:
        A CVResults tuple with the following fields

        - model: GPyTorchModel for batched cross validation
        - posterior: GPyTorchPosterior where the mean has shape `n x 1 x m` or
          `batch_shape x n x 1 x m`
        - observed_Y: A `n x 1 x m` or `batch_shape x n x 1 x m` tensor of observations.
        - observed_Yvar: A `n x 1 x m` or `batch_shape x n x 1 x m` tensor of observed
          measurement noise.

    Example:
        >>> import torch
        >>> from botorch.cross_validation import (
        ...     batch_cross_validation, gen_loo_cv_folds
        ... )
        >>>
        >>> from botorch.models import SingleTaskGP
        >>> from botorch.models.transforms.input import Normalize
        >>> from botorch.models.transforms.outcome import Standardize
        >>> from gpytorch.mlls import ExactMarginalLogLikelihood

        >>> train_X = torch.rand(10, 1)
        >>> train_Y = torch.rand_like(train_X)
        >>> cv_folds = gen_loo_cv_folds(train_X, train_Y)
        >>> input_transform = Normalize(d=train_X.shape[-1])
        >>>
        >>> cv_results = batch_cross_validation(
        ...    model_cls=SingleTaskGP,
        ...    mll_cls=ExactMarginalLogLikelihood,
        ...    cv_folds=cv_folds,
        ...    model_init_kwargs={
        ...        "input_transform": input_transform,
        ...    },
        ... )
    """
    if issubclass(model_cls, MultiTaskGP):
        raise UnsupportedError(
            "Multi-task GPs are not currently supported by `batch_cross_validation`."
        )
    model_init_kws = model_init_kwargs if model_init_kwargs is not None else {}
    if cv_folds.train_Yvar is not None:
        model_init_kws["train_Yvar"] = cv_folds.train_Yvar
    model_cv = model_cls(
        train_X=cv_folds.train_X,
        train_Y=cv_folds.train_Y,
        **model_init_kws,
    )
    mll_cv = mll_cls(model_cv.likelihood, model_cv)
    mll_cv.to(cv_folds.train_X)

    fit_args = fit_args or {}
    mll_cv = fit_gpytorch_mll(mll_cv, **fit_args)

    # Evaluate on the hold-out set in batch mode
    with torch.no_grad():
        posterior = model_cv.posterior(
            cv_folds.test_X, observation_noise=observation_noise
        )

    return CVResults(
        model=model_cv,
        posterior=posterior,
        observed_Y=cv_folds.test_Y,
        observed_Yvar=cv_folds.test_Yvar,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import torch
from botorch.models import SingleTaskGP
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.testing import BotorchTestCase, MockModel, MockPosterior
from botorch_community.acquisition.hentropy_search import (
    qHEntropySearch,
    qLossFunctionMinMax,
    qLossFunctionTopK,
)

NO = "botorch.utils.testing.MockModel.num_outputs"


class TestQHEntropySearch(BotorchTestCase):
    def test_initialize_q_hentropy_search(self):
        for dtype in (torch.float, torch.double):
            mean = torch.zeros(1, 1, device=self.device, dtype=dtype)
            mm = MockModel(MockPosterior(mean=mean))
            # test error when neither specifying neither sampler nor num_fantasies
            with self.assertRaisesRegex(
                ValueError, "Must specify `num_points` if no `sampler` is provided."
            ):
                qHEntropySearch(
                    model=mm,
                    loss_function_class=qLossFunctionMinMax,
                    loss_function_hyperparameters={},
                    n_fantasy_at_design_pts=None,
                    n_fantasy_at_action_pts=None,
                )
            # test error when sampler and num_fantasies arg are inconsistent
            design_sampler = SobolQMCNormalSampler(sample_shape=torch.Size([16]))
            action_sampler = SobolQMCNormalSampler(sample_shape=torch.Size([16]))
            with self.assertRaises(ValueError):
                qHEntropySearch(
                    model=mm,
                    loss_function_class=qLossFunctionMinMax,
                    loss_function_hyperparameters={},
                    n_fantasy_at_design_pts=32,
                    n_fantasy_at_action_pts=32,
                    design_sampler=design_sampler,
                    action_sampler=action_sampler,
                )

            # test default construction
            hes = qHEntropySearch(
                model=mm,
                loss_function_class=qLossFunctionMinMax,
                loss_function_hyperparameters={},
            )
            self.assertEqual(hes.n_fantasy_at_design_pts, 64)
            self.assertEqual(hes.n_fantasy_at_action_pts, 64)
            self.assertIsInstance(hes.design_sampler, SobolQMCNormalSampler)
            self.assertIsInstance(hes.action_sampler, SobolQMCNormalSampler)
            self.assertEqual(hes.design_sampler.sample_shape, torch.Size([64]))
            self.assertEqual(hes.action_sampler.sample_shape, torch.Size([64]))
            self.assertEqual(hes.get_augmented_q_batch_size(q=3), 64 + 3)

            # test custom construction
            design_sampler = SobolQMCNormalSampler(sample_shape=torch.Size([16]))
            action_sampler = SobolQMCNormalSampler(sample_shape=torch.Size([16]))
            hes = qHEntropySearch(
                model=mm,
                loss_function_class=qLossFunctionMinMax,
                loss_function_hyperparameters={},
                n_fantasy_at_design_pts=16,
                n_fantasy_at_action_pts=16,
                design_sampler=design_sampler,
                action_sampler=action_sampler,
            )
            self.assertEqual(hes.n_fantasy_at_design_pts, 16)
            self.assertEqual(hes.n_fantasy_at_action_pts, 16)
            self.assertIsInstance(hes.design_sampler, SobolQMCNormalSampler)
            self.assertIsInstance(hes.action_sampler, SobolQMCNormalSampler)
            self.assertEqual(hes.design_sampler.sample_shape, torch.Size([16]))
            self.assertEqual(hes.action_sampler.sample_shape, torch.Size([16]))
            self.assertEqual(hes.get_augmented_q_batch_size(q=3), 16 + 3)

            # test assignment of num_fantasies from sampler if not provided
            hes = qHEntropySearch(
                model=mm,
                loss_function_class=qLossFunctionMinMax,
                loss_function_hyperparameters={},
                n_fantasy_at_design_pts=None,
                n_fantasy_at_action_pts=None,
                design_sampler=design_sampler,
                action_sampler=action_sampler,
            )
            self.assertEqual(hes.design_sampler.sample_shape, torch.Size([16]))
            self.assertEqual(hes.action_sampler.sample_shape, torch.Size([16]))

    def test_qhentropy_search(self):
        d = 2
        q = 2
        num_data = 3
        num_fantasies = 4
        t_batch_size = 8

        # MinMax loss function
        for dtype in (torch.float, torch.double):
            bounds = torch.tensor([[0], [1]], device=self.device, dtype=dtype)
            bounds = bounds.repeat(1, d)
            train_X = torch.rand(num_data, d, device=self.device, dtype=dtype)
            train_Y = torch.rand(num_data, 1, device=self.device, dtype=dtype)
            model = SingleTaskGP(train_X, train_Y)

            # default evaluation tests
            hes = qHEntropySearch(
                model=model,
                loss_function_class=qLossFunctionMinMax,
                loss_function_hyperparameters={},
                n_fantasy_at_design_pts=num_fantasies,
                n_fantasy_at_action_pts=num_fantasies,
            )
            eval_X = torch.rand(t_batch_size, q, d, device=self.device, dtype=dtype)
            eval_A = torch.rand(
                t_batch_size, num_fantasies, q, d, device=self.device, dtype=dtype
            )
            result = hes(X=eval_X, A=eval_A)
            self.assertEqual(result.shape, torch.Size([t_batch_size]))

            # add dummy base_weights to samplers
            samplers = [
                SobolQMCNormalSampler(sample_shape=torch.Size([num_fantasies]))
                for _ in range(2)
            ]
            for s in samplers:
                s.base_weights = torch.ones(
                    s.sample_shape[0], 1, device=self.device, dtype=dtype
                )

            hes = qHEntropySearch(
                model=model,
                loss_function_class=qLossFunctionMinMax,
                loss_function_hyperparameters={},
                n_fantasy_at_design_pts=num_fantasies,
                n_fantasy_at_action_pts=num_fantasies,
                design_sampler=samplers[0],
                action_sampler=samplers[1],
            )
            q_prime = hes.get_augmented_q_batch_size(q)

            eval_X = torch.rand(
                t_batch_size, q_prime, d, device=self.device, dtype=dtype
            )
            eval_A = torch.rand(
                t_batch_size, num_fantasies, q, d, device=self.device, dtype=dtype
            )
            result = hes(X=eval_X, A=eval_A)
            self.assertEqual(result.shape, torch.Size([t_batch_size]))

            # extract candidates
            cand = hes.extract_candidates(eval_X)
            self.assertEqual(cand.shape, torch.Size([t_batch_size, q, d]))

        # TopK loss function
        for dtype in (torch.float, torch.double):
            bounds = torch.tensor([[0], [1]], device=self.device, dtype=dtype)
            bounds = bounds.repeat(1, d)
            train_X = torch.rand(num_data, d, device=self.device, dtype=dtype)
            train_Y = torch.rand(num_data, 1, device=self.device, dtype=dtype)
            model = SingleTaskGP(train_X, train_Y)

            # default evaluation tests
            hes = qHEntropySearch(
                model=model,
                loss_function_class=qLossFunctionTopK,
                loss_function_hyperparameters={
                    "dist_weight": 1.0,
                    "dist_threshold": 0.5,
                },
                n_fantasy_at_design_pts=num_fantasies,
                n_fantasy_at_action_pts=num_fantasies,
            )
            eval_X = torch.rand(t_batch_size, q, d, device=self.device, dtype=dtype)
            eval_A = torch.rand(
                t_batch_size, num_fantasies, q, d, device=self.device, dtype=dtype
            )
            result = hes(X=eval_X, A=eval_A)
            self.assertEqual(result.shape, torch.Size([t_batch_size]))

            # add dummy base_weights to samplers
            samplers = [
                SobolQMCNormalSampler(sample_shape=torch.Size([num_fantasies]))
                for _ in range(2)
            ]
            for s in samplers:
                s.base_weights = torch.ones(
                    s.sample_shape[0], 1, device=self.device, dtype=dtype
                )

            hes = qHEntropySearch(
                model=model,
                loss_function_class=qLossFunctionTopK,
                loss_function_hyperparameters={
                    "dist_weight": 1.0,
                    "dist_threshold": 0.5,
                },
                n_fantasy_at_design_pts=num_fantasies,
                n_fantasy_at_action_pts=num_fantasies,
                design_sampler=samplers[0],
                action_sampler=samplers[1],
            )
            q_prime = hes.get_augmented_q_batch_size(q)

            eval_X = torch.rand(
                t_batch_size, q_prime, d, device=self.device, dtype=dtype
            )
            eval_A = torch.rand(
                t_batch_size, num_fantasies, q, d, device=self.device, dtype=dtype
            )
            result = hes(X=eval_X, A=eval_A)
            self.assertEqual(result.shape, torch.Size([t_batch_size]))

            # extract candidates
            cand = hes.extract_candidates(eval_X)
            self.assertEqual(cand.shape, torch.Size([t_batch_size, q, d]))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import torch
from botorch.logging import logger
from botorch.posteriors.ensemble import EnsemblePosterior
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.posterior import Posterior
from botorch.posteriors.posterior_list import PosteriorList
from botorch.posteriors.torch import TorchPosterior
from botorch.posteriors.transformed import TransformedPosterior
from botorch.sampling.base import MCSampler
from botorch.sampling.index_sampler import IndexSampler
from botorch.sampling.list_sampler import ListSampler
from botorch.sampling.normal import (
    IIDNormalSampler,
    NormalMCSampler,
    SobolQMCNormalSampler,
)
from botorch.utils.dispatcher import Dispatcher
from gpytorch.distributions import MultivariateNormal
from torch.distributions import Distribution
from torch.quasirandom import SobolEngine


def _posterior_to_distribution_encoder(
    posterior: Posterior,
) -> type[Distribution] | type[Posterior]:
    r"""An encoder returning the type of the distribution for `TorchPosterior`
    and the type of the posterior for the rest.
    """
    if isinstance(posterior, TorchPosterior):
        return type(posterior.distribution)
    return type(posterior)


GetSampler = Dispatcher("get_sampler", encoder=_posterior_to_distribution_encoder)


def get_sampler(
    posterior: TorchPosterior,
    sample_shape: torch.Size,
    *,
    seed: int | None = None,
) -> MCSampler:
    r"""Get the sampler for the given posterior.

    The sampler can be used as `sampler(posterior)` to produce samples
    suitable for use in acquisition function optimization via SAA.

    Args:
        posterior: A `Posterior` to get the sampler for.
        sample_shape: The sample shape of the samples produced by the
            given sampler. The full shape of the resulting samples is
            given by `posterior._extended_shape(sample_shape)`.
        seed: Seed used to initialize sampler.

    Returns:
        The `MCSampler` object for the given posterior.
    """
    return GetSampler(posterior, sample_shape=sample_shape, seed=seed)


@GetSampler.register(MultivariateNormal)
def _get_sampler_mvn(
    posterior: GPyTorchPosterior,
    sample_shape: torch.Size,
    *,
    seed: int | None = None,
) -> NormalMCSampler:
    r"""The Sobol normal sampler for the `MultivariateNormal` posterior.

    If the output dim is too large, falls back to `IIDNormalSampler`.
    """
    sampler = SobolQMCNormalSampler(sample_shape=sample_shape, seed=seed)
    collapsed_shape = sampler._get_collapsed_shape(posterior=posterior)
    base_collapsed_shape = collapsed_shape[len(sample_shape) :]
    if base_collapsed_shape.numel() > SobolEngine.MAXDIM:
        logger.warning(
            f"Output dim {base_collapsed_shape.numel()} is too large for the "
            "Sobol engine. Using IIDNormalSampler instead."
        )
        sampler = IIDNormalSampler(sample_shape=sample_shape, seed=seed)
    return sampler


@GetSampler.register(TransformedPosterior)
def _get_sampler_derived(
    posterior: TransformedPosterior,
    sample_shape: torch.Size,
    *,
    seed: int | None = None,
) -> MCSampler:
    r"""Get the sampler for the underlying posterior."""
    return get_sampler(
        posterior=posterior._posterior,
        sample_shape=sample_shape,
        seed=seed,
    )


@GetSampler.register(PosteriorList)
def _get_sampler_list(
    posterior: PosteriorList, sample_shape: torch.Size, *, seed: int | None = None
) -> MCSampler:
    r"""Get the `ListSampler` with the appropriate list of samplers.

    NOTE: Does not dispatch to Sobol sampling for normal posteriors due to
    correlations between samplers. Instead uses `IIDNormalSampler`. See the following
    issue for details: https://github.com/meta-pytorch/botorch/issues/2658
    """
    samplers = []
    for p in posterior.posteriors:
        sampler = get_sampler(posterior=p, sample_shape=sample_shape, seed=seed)
        if isinstance(sampler, SobolQMCNormalSampler):
            sampler = IIDNormalSampler(sample_shape=sample_shape, seed=seed)
        samplers.append(sampler)

    return ListSampler(*samplers)


@GetSampler.register(EnsemblePosterior)
def _get_sampler_ensemble(
    posterior: EnsemblePosterior,
    sample_shape: torch.Size,
    seed: int | None = None,
) -> MCSampler:
    r"""Get the `IndexSampler` for the `EnsemblePosterior`."""
    return IndexSampler(sample_shape=sample_shape, seed=seed)


@GetSampler.register(object)
def _not_found_error(
    posterior: Posterior,
    sample_shape: torch.Size,
    seed: int | None = None,
) -> None:
    raise NotImplementedError(
        f"A registered `MCSampler` for posterior {posterior} is not found. You can "
        "implement and register one using `@GetSampler.register`."
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Linear Elliptical Slice Sampler.

References

.. [Gessner2020]
    A. Gessner, O. Kanjilal, and P. Hennig. Integrals over gaussians under
    linear domain constraints. AISTATS 2020.

.. [Wu2024]
    K. Wu, and J. Gardner. A Fast, Robust Elliptical Slice Sampling Implementation for
    Linearly Truncated Multivariate Normal Distributions. arXiv:2407.10449. 2024.

This implementation is based (with multiple changes / optimiations) on
the following implementations based on the algorithm in [Gessner2020]_:
- https://github.com/alpiges/LinConGauss
- https://github.com/wjmaddox/pytorch_ess

In addition, the active intervals (from which the angle is sampled) are computed using
the improved algorithm described in [Wu2024]_:
https://github.com/kayween/linear-ess

The implementation here differentiates itself from the original implementations with:
1) Support for fixed feature equality constraints.
2) Support for non-standard Normal distributions.
3) Numerical stability improvements, especially relevant for high-dimensional cases.
4) Support multiple Markov chains running in parallel.
"""

from __future__ import annotations

import math

import torch
from botorch.utils.sampling import PolytopeSampler
from linear_operator.operators import DiagLinearOperator, LinearOperator
from torch import Tensor

_twopi = 2.0 * math.pi


class LinearEllipticalSliceSampler(PolytopeSampler):
    r"""Linear Elliptical Slice Sampler.

    Ideas:
    - Optimize computations if possible, potentially with torch.compile.
    - Extend fixed features constraint to general linear equality constraints.
    """

    def __init__(
        self,
        inequality_constraints: tuple[Tensor, Tensor] | None = None,
        bounds: Tensor | None = None,
        interior_point: Tensor | None = None,
        fixed_indices: list[int] | Tensor | None = None,
        mean: Tensor | None = None,
        covariance_matrix: Tensor | LinearOperator | None = None,
        covariance_root: Tensor | LinearOperator | None = None,
        check_feasibility: bool = False,
        burnin: int = 0,
        thinning: int = 0,
        num_chains: int = 1,
    ) -> None:
        r"""Initialize LinearEllipticalSliceSampler.

        Args:
            inequality_constraints: Tensors `(A, b)` describing inequality constraints
                 `A @ x <= b`, where `A` is an `n_ineq_con x d`-dim Tensor and `b` is
                 an `n_ineq_con x 1`-dim Tensor, with `n_ineq_con` the number of
                 inequalities and `d` the dimension of the sample space. If omitted,
                 must provide `bounds` instead.
            bounds: A `2 x d`-dim tensor of box bounds. If omitted, must provide
                `inequality_constraints` instead.
            interior_point: A `d x 1`-dim Tensor presenting a point in the (relative)
                interior of the polytope. If omitted, an interior point is determined
                automatically by solving a Linear Program. Note: It is crucial that
                the point lie in the interior of the feasible set (rather than on the
                boundary), otherwise the sampler will produce invalid samples.
            fixed_indices: Integer list or `d`-dim Tensor representing the indices of
                dimensions that are constrained to be fixed to the values specified in
                the `interior_point`, which is required to be passed in conjunction with
                `fixed_indices`.
            mean: The `d x 1`-dim mean of the MVN distribution (if omitted, use zero).
            covariance_matrix: The `d x d`-dim covariance matrix of the MVN
                distribution (if omitted, use the identity).
            covariance_root: A `d x d`-dim root of the covariance matrix such that
                covariance_root @ covariance_root.T = covariance_matrix. NOTE: This
                matrix is assumed to be lower triangular. covariance_root can only be
                passed in conjunction with fixed_indices if covariance_root is a
                DiagLinearOperator. Otherwise the factorization would need to be re-
                computed, as we need to solve in `standardize`.
            check_feasibility: If True, raise an error if the sampling results in an
                infeasible sample. This creates some overhead and so is switched off
                by default.
            burnin: Number of samples to generate upon initialization to warm up the
                sampler.
            thinning: Number of samples to skip before returning a sample in `draw`.
            num_chains: Number of Markov chains to run in parallel.

        This sampler samples from a multivariante Normal `N(mean, covariance_matrix)`
        subject to linear domain constraints `A x <= b` (intersected with box bounds,
        if provided).
        """
        if interior_point is not None and interior_point.ndim == 1:
            interior_point = interior_point.unsqueeze(-1)

        if mean is not None and mean.ndim == 1:
            mean = mean.unsqueeze(-1)

        super().__init__(
            inequality_constraints=inequality_constraints,
            # TODO: Support equality constraints?
            interior_point=interior_point,
            bounds=bounds,
        )
        tkwargs = {"device": self.x0.device, "dtype": self.x0.dtype}
        if covariance_matrix is not None and covariance_root is not None:
            raise ValueError(
                "Provide either covariance_matrix or covariance_root, not both."
            )

        # can't unpack inequality constraints directly if bounds are passed
        A, b = self.A, self.b
        self._Az, self._bz = A, b
        self._is_fixed, self._not_fixed = None, None
        if fixed_indices is not None:
            mean, covariance_matrix, covariance_root = (
                self._fixed_features_initialization(
                    A=A,
                    b=b,
                    interior_point=interior_point,
                    fixed_indices=fixed_indices,
                    mean=mean,
                    covariance_matrix=covariance_matrix,
                    covariance_root=covariance_root,
                )
            )

        self._mean = mean
        # Have to delay factorization until after fixed features initialization.
        if covariance_matrix is not None:  # implies root is None
            covariance_root, info = torch.linalg.cholesky_ex(covariance_matrix)
            not_psd = torch.any(info)
            if not_psd:
                raise ValueError(
                    "Covariance matrix is not positive definite. "
                    "Currently only non-degenerate distributions are supported."
                )
        self._covariance_root = covariance_root

        # Rewrite the constraints as a system that constrains a standard Normal.
        self._standardization_initialization()

        # state of the sampler ("current point")
        self._x = self.x0.clone()
        self._z = self._transform(self._x)

        # Expand the shape to (d, num_chains) for running parallel Markov chains.
        if num_chains > 1:
            self._z = self._z.expand(-1, num_chains).clone()

        # We will need the following repeatedly, let's allocate them once
        self.zeros = torch.zeros((num_chains, 1), **tkwargs)
        self.ones = torch.ones((num_chains, 1), **tkwargs)
        self.indices_batch = torch.arange(
            num_chains, dtype=torch.int64, device=tkwargs["device"]
        )

        self.check_feasibility = check_feasibility
        self._lifetime_samples = 0
        if burnin > 0:
            self.thinning = 0
            self.draw(burnin)
        self.thinning = thinning

    def _fixed_features_initialization(
        self,
        A: Tensor,
        b: Tensor,
        interior_point: Tensor | None,
        fixed_indices: list[int] | Tensor,
        mean: Tensor | None,
        covariance_matrix: Tensor | None,
        covariance_root: Tensor | None,
    ) -> tuple[Tensor | None, Tensor | None]:
        """Modifies the constraint system (A, b) due to fixed indices and assigns
        the modified constraints system to `self._Az`, `self._bz`. NOTE: Needs to be
        called prior to `self._standardization_initialization` in the constructor.
        covariance_root and fixed_indices can both not be None only if covariance_root
        is a DiagLinearOperator. Otherwise, the covariance matrix would need to be
        refactorized.

        Returns:
            Tuple of `mean` and `covariance_matrix` tensors of the non-fixed dimensions.
        """
        if interior_point is None:
            raise ValueError(
                "If `fixed_indices` are provided, an interior point must also be "
                "provided in order to infer feasible values of the fixed features."
            )

        root_is_diag = isinstance(covariance_root, DiagLinearOperator)
        if covariance_root is not None and not root_is_diag:
            root_is_diag = (covariance_root.diag().diag() == covariance_root).all()
            if root_is_diag:  # convert the diagonal root to a DiagLinearOperator
                covariance_root = DiagLinearOperator(covariance_root.diagonal())
            else:  # otherwise, fail
                raise ValueError(
                    "Provide either covariance_root or fixed_indices, not both."
                )
        d = interior_point.shape[0]
        is_fixed, not_fixed = get_index_tensors(fixed_indices=fixed_indices, d=d)
        self._is_fixed = is_fixed
        self._not_fixed = not_fixed
        # Transforming constraint system to incorporate fixed features:
        # A @ x - b = (A[:, fixed] @ x[fixed] + A[:, not fixed] @ x[not fixed]) - b
        #           = A[:, not fixed] @ x[not fixed] - (b - A[:, fixed] @ x[fixed])
        #           = Az @ z - bz
        self._Az = A[:, not_fixed]
        self._bz = b - A[:, is_fixed] @ interior_point[is_fixed]
        if mean is not None:
            mean = mean[not_fixed]
        if covariance_matrix is not None:  # subselect active dimensions
            covariance_matrix = covariance_matrix[
                not_fixed.unsqueeze(-1), not_fixed.unsqueeze(0)
            ]
        if root_is_diag:  # in the special case of diagonal root, can subselect
            covariance_root = DiagLinearOperator(covariance_root.diagonal()[not_fixed])

        return mean, covariance_matrix, covariance_root

    def _standardization_initialization(self) -> None:
        """For non-standard mean and covariance, we're going to rewrite the problem as
        sampling from a standard normal distribution subject to modified constraints.
            A @ x - b = A @ (covar_root @ z + mean) - b
                      = (A @ covar_root) @ z - (b - A @ mean)
                      = _Az @ z - _bz
        NOTE: We need to standardize bz before Az in the following, because it relies
        on the untransformed Az. We can't simply use A instead because Az might have
        been subject to the fixed features transformation.
        """
        if self._mean is not None:
            self._bz = self._bz - self._Az @ self._mean
        if self._covariance_root is not None:
            self._Az = self._Az @ self._covariance_root

    @property
    def lifetime_samples(self) -> int:
        """The total number of samples generated by the sampler during its lifetime."""
        return self._lifetime_samples

    def draw(self, n: int = 1) -> Tensor:
        r"""Draw samples.

        Args:
            n: The number of samples.

        Returns:
            A `(n * num_chains) x d`-dim tensor of `n * num_chains` samples.
        """
        samples = []
        for _ in range(n):
            for _ in range(self.thinning):
                self.step()
            samples.append(self.step())
        return torch.cat(samples, dim=-1).transpose(-1, -2)

    def step(self) -> Tensor:
        r"""Take a step, return the new sample, update the internal state.

        Returns:
            A `d x num_chains`-dim tensor, where each column is a sample from a Markov
            chain.
        """
        nu = torch.randn_like(self._z)
        theta = self._draw_angle(nu=nu)

        self._z = z = self._get_cart_coords(nu=nu, theta=theta)
        self._x = x = self._untransform(z)

        self._lifetime_samples += 1
        if self.check_feasibility and (not self._is_feasible(self._x).all()):
            Axmb = self.A @ self._x - self.b
            violated_indices = Axmb > 0
            raise RuntimeError(
                "Sampling resulted in infeasible point. \n\t- Number "
                f"of violated constraints: {violated_indices.sum()}."
                f"\n\t- Magnitude of violations: {Axmb[violated_indices]}"
                "\n\t- If the error persists, please report this bug on GitHub."
            )
        return x

    def _draw_angle(self, nu: Tensor) -> Tensor:
        r"""Draw the rotation angle.

        Args:
            nu: A `d x num_chains`-dim tensor (the "new" direction, drawn from N(0, I)).

        Returns:
            A `num_chains`-dim Tensor containing the rotation angle (radians).
        """
        left, right = self._find_active_intersection_angles(nu)
        left, right = self._trim_intervals(left, right)

        # If left[i, j] <= right[i, j], then [left[i, j], right[i, j]] is an active
        # interval. On the other hand, if left[i, j] > right[i, j], then they are both
        # dummy variables and should be discarded. Thus, we clamp their difference so
        # that they do not contribute to the cumulative length.
        csum = right.sub(left).clamp(min=0.0).cumsum(dim=-1)

        u = csum[:, -1] * torch.rand(
            right.size(-2), dtype=right.dtype, device=right.device
        )

        # The returned index i satisfies csum[i - 1] < u <= csum[i]
        idx = torch.searchsorted(csum, u.unsqueeze(-1)).squeeze(-1)

        # Do a zero padding so that padded_csum[i] = csum[i - 1]
        padded_csum = torch.cat([self.zeros, csum], dim=-1)

        return u - padded_csum[self.indices_batch, idx] + left[self.indices_batch, idx]

    def _get_cart_coords(self, nu: Tensor, theta: Tensor) -> Tensor:
        r"""Determine location on the ellipse in Cartesian coordinates.

        Args:
            nu: A `d x num_chains`-dim tensor (the "new" direction, drawn from N(0, I)).
            theta: A `num_chains`-dim tensor of angles.

        Returns:
            A `d x num_chains`-dim tensor of samples from the domain in Cartesian
            coordinates.
        """
        return self._z * torch.cos(theta) + nu * torch.sin(theta)

    def _trim_intervals(self, left: Tensor, right: Tensor) -> tuple[Tensor, Tensor]:
        """Trim the intervals by a small positive constant. This encourages the Markov
        chain to stay in the interior of the domain.
        """
        gap = torch.clamp(right - left, min=0.0)
        eps = gap.mul(0.25).clamp(max=1e-6 if gap.dtype == torch.float32 else 1e-12)

        return left + eps, right - eps

    def _find_active_intersection_angles(self, nu: Tensor) -> tuple[Tensor, Tensor]:
        """Construct the active intersection angles.

        Args:
            nu: A `d x num_chains`-dim tensor (the "new" direction, drawn from N(0, I)).

        Returns:
            A tuple (left, right) of two tensors of size `num_chains x m` representing
            the active intersection angles. For the i-th Markov chain and the j-th
            constraint, a pair of angles left[i, j] and right[i, j] is active if and
            only if left[i, j] <= right[i, j]. If left[i, j] > right[i, j], they are
            inactive and should be ignored.
        """
        alpha, beta = self._find_intersection_angles(nu)

        # It's easier to put `num_chains` as the first dimension,
        # because `torch.searchsorted` only supports searching in the last dimension
        alpha, beta = alpha.T, beta.T

        srted, indices = torch.sort(alpha, descending=False)
        cummax = beta[self.indices_batch.unsqueeze(-1), indices].cummax(dim=-1).values

        srted = torch.cat([srted, self.ones * 2 * math.pi], dim=-1)
        cummax = torch.cat([self.zeros, cummax], dim=-1)

        return cummax, srted

    def _find_intersection_angles(self, nu: Tensor) -> tuple[Tensor, Tensor]:
        """Compute all 2 * m intersections of the ellipse and the domain, where
        `m = n_ineq_con` is the number of inequality constraints defining the domain.
        If the i-th linear inequality constraint has no intersection with the ellipse,
        we will create two dummy intersection angles alpha_i = beta_i = 0.

        Args:
            nu: A `d x num_chains`-dim tensor (the "new" direction, drawn from N(0, I)).

        Returns:
            A tuple of two tensors with the same size `m x num_chains`. The first tensor
            represents the smaller intersection angles. The second tensor represents the
            larger intersection angles.
        """
        p = self._Az @ self._z
        q = self._Az @ nu

        radius = torch.sqrt(p**2 + q**2)

        ratio = self._bz / radius

        has_solution = ratio < 1.0

        arccos = torch.arccos(ratio)
        arccos[~has_solution] = 0.0
        arctan = torch.arctan2(q, p)

        theta1 = arctan + arccos
        theta2 = arctan - arccos

        # translate every angle to [0, 2 * pi]
        theta1 = theta1 + theta1.lt(0.0) * _twopi
        theta2 = theta2 + theta2.lt(0.0) * _twopi

        alpha = torch.minimum(theta1, theta2)
        beta = torch.maximum(theta1, theta2)

        return alpha, beta

    def _is_feasible(self, points: Tensor, transformed: bool = False) -> Tensor:
        r"""Returns a Boolean tensor indicating whether the `points` are feasible,
        i.e. they satisfy `A @ points <= b`, where `(A, b)` are the tensors passed
        as the `inequality_constraints` to the constructor of the sampler.

        Args:
            points: A `d x M`-dim tensor of points.
            transformed: Wether points are assumed to be transformed by a change of
                basis, which means feasibility should be computed based on the
                transformed constraint system (_Az, _bz), instead of (A, b).

        Returns:
            An `M`-dim binary tensor where `True` indicates that the associated
            point is feasible.
        """
        A, b = (self._Az, self._bz) if transformed else (self.A, self.b)
        return (A @ points <= b).all(dim=0)

    def _transform(self, x: Tensor) -> Tensor:
        """Transforms the input so that it is equivalent to a standard Normal variable
        constrained with the modified system constraints (self._Az, self._bz).

        Args:
            x: The input tensor to be transformed, usually `d x 1`-dimensional.

        Returns:
            A `d x 1`-dimensional tensor of transformed values subject to the modified
            system of constraints.
        """
        if self._not_fixed is not None:
            x = x[self._not_fixed]
        return self._standardize(x)

    def _untransform(self, z: Tensor) -> Tensor:
        """The inverse transform of the `_transform`, i.e. maps `z` back to the original
        space where it is subject to the original constraint system (self.A, self.b).

        Args:
            z: The transformed tensor to be un-transformed, usually `d x 1`-dimensional.

        Returns:
            A `d x 1`-dimensional tensor of un-transformed values subject to the
            original system of constraints.
        """
        if self._is_fixed is None:
            return self._unstandardize(z)
        else:
            x = self._x.clone()  # _x already contains the fixed values
            x[self._not_fixed] = self._unstandardize(z)
            return x

    def _standardize(self, x: Tensor) -> Tensor:
        """_transform helper standardizing the input `x`, which is assumed to be a
        `d x 1`-dim Tensor, or a `len(self._not_fixed) x 1`-dim if there are fixed
        indices.
        """
        z = x
        if self._mean is not None:
            z = z - self._mean
        root = self._covariance_root
        if root is not None:
            z = torch.linalg.solve_triangular(root, z, upper=False)

        return z

    def _unstandardize(self, z: Tensor) -> Tensor:
        """_untransform helper un-standardizing the input `z`, which is assumed to be a
        `d x 1`-dim Tensor, or a `len(self._not_fixed) x 1`-dim if there are fixed
        indices.
        """
        x = z
        if self._covariance_root is not None:
            x = self._covariance_root @ x
        if self._mean is not None:
            x = x + self._mean
        return x


def get_index_tensors(
    fixed_indices: list[int] | Tensor, d: int
) -> tuple[Tensor, Tensor]:
    """Converts `fixed_indices` to a `d`-dim integral Tensor that is True at indices
    that are contained in `fixed_indices` and False otherwise.

    Args:
        fixed_indices: A list or Tensoro of integer indices to fix.
        d: The dimensionality of the Tensors to be indexed.

    Returns:
        A Tuple of integral Tensors partitioning [1, d] into indices that are fixed
        (first tensor) and non-fixed (second tensor).
    """
    is_fixed = torch.as_tensor(fixed_indices)
    dtype, device = is_fixed.dtype, is_fixed.device
    dims = torch.arange(d, dtype=dtype, device=device)
    not_fixed = torch.tensor([i for i in dims if i not in is_fixed], dtype=dtype)
    return is_fixed, not_fixed



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import time
import warnings
from collections.abc import Callable, Sequence
from typing import Any

import numpy.typing as npt
from botorch.exceptions.errors import OptimizationTimeoutError
from scipy import optimize
from threadpoolctl import threadpool_limits


def minimize_with_timeout(
    fun: Callable[[npt.NDArray, ...], float],
    x0: npt.NDArray,
    args: tuple[Any, ...] = (),
    method: str | None = None,
    jac: str | Callable | bool | None = None,
    hess: str | Callable | optimize.HessianUpdateStrategy | None = None,
    hessp: Callable | None = None,
    bounds: Sequence[tuple[float, float]] | optimize.Bounds | None = None,
    constraints=(),  # Typing this properly is a s**t job
    tol: float | None = None,
    callback: Callable | None = None,
    options: dict[str, Any] | None = None,
    timeout_sec: float | None = None,
) -> optimize.OptimizeResult:
    r"""Wrapper around scipy.optimize.minimize to support timeout.

    This method calls scipy.optimize.minimize with all arguments forwarded
    verbatim. The only difference is that if provided a `timeout_sec` argument,
    it will automatically stop the optimization after the timeout is reached.

    Internally, this is achieved by automatically constructing a wrapper callback
    method that is injected to the scipy.optimize.minimize call and that keeps
    track of the runtime and the optimization variables at the current iteration.
    """
    if timeout_sec is not None:
        start_time = time.monotonic()
        callback_data = {"num_iterations": 0}  # update from withing callback below

        def timeout_callback(xk: npt.NDArray) -> bool:
            runtime = time.monotonic() - start_time
            callback_data["num_iterations"] += 1
            if runtime > timeout_sec:
                raise OptimizationTimeoutError(current_x=xk, runtime=runtime)
            return False

        if callback is None:
            wrapped_callback = timeout_callback

        elif callable(method):
            raise NotImplementedError(
                "Custom callable not supported for `method` argument."
            )

        elif method == "trust-constr":  # special signature

            def wrapped_callback(
                xk: npt.NDArray, state: optimize.OptimizeResult
            ) -> bool:
                # order here is important to make sure base callback gets executed
                return callback(xk, state) or timeout_callback(xk=xk)

        else:

            def wrapped_callback(xk: npt.NDArray) -> None:
                timeout_callback(xk=xk)
                callback(xk)

    else:
        wrapped_callback = callback

    try:
        warnings.filterwarnings("error", message="Method .* cannot handle")
        # To prevent slowdowns after scipy 1.15.
        # See https://github.com/scipy/scipy/issues/22438.
        with threadpool_limits(limits=1, user_api="blas"):
            return optimize.minimize(
                fun=fun,
                x0=x0,
                args=args,
                method=method,
                jac=jac,
                hess=hess,
                hessp=hessp,
                bounds=bounds,
                constraints=constraints,
                tol=tol,
                callback=wrapped_callback,
                options=options,
            )
    except OptimizationTimeoutError as e:
        msg = f"Optimization timed out after {e.runtime} seconds."
        current_fun, *_ = fun(e.current_x, *args)

        return optimize.OptimizeResult(
            fun=current_fun,
            x=e.current_x,
            nit=callback_data["num_iterations"],
            success=False,  # same as when maxiter is reached
            status=1,  # same as when L-BFGS-B reaches maxiter
            message=msg,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Region Averaged Acquisition Functions for Efficient Trust Region Selection
for High-Dimensional Trust Region Bayesian Optimization.
See [eriksson2019TuRBO]_, [namura2024rei]_

Two versions of the Regional Expected Improvement (REI) acquisition
function are implemented here:
1. Analytic version: LogRegionalExpectedImprovement
2. Monte Carlo version: qLogRegionalExpectedImprovement

LogREI has been implemented from the original paper [namura2024rei]_
and code(https://github.com/Nobuo-Namura/regional-expected-improvement/)
and qLogREI have been implemented by slightly modifying the qREI
acquisition function implementation from the paper.

These acquisition functions can help explore the design space efficiently
since trust regions at initialization and restarts in algorithms like TuRBO
are selected by optimizing the Region Averaged Acquisition Functions instead
of sampling them randomly. This has displayed faster convergence in some cases
and convergence to better solutions in general as showed in [namura2024rei]_.

References

.. [eriksson2019TuRBO]
    D. Eriksson, M. Pearce, J.R. Gardner, R. Turner, M. Poloczek.
    Scalable Global Optimization via Local Bayesian Optimization.
    Advances in Neural Information Processing Systems, 2019.
.. [namura2024rei]
    Nobuo Namura, Sho Takemori.
    Regional Expected Improvement for Efficient Trust Region
    Selection in High-Dimensional Bayesian Optimization.
    Proceedings of the 39th AAAI Conference on Artificial
    Intelligence, 2025.

Contributor: SaiAakash
"""

from __future__ import annotations

from functools import partial

import torch
from botorch.acquisition.analytic import (
    _log_ei_helper,
    _scaled_improvement,
    AnalyticAcquisitionFunction,
)
from botorch.acquisition.logei import (
    _log_improvement,
    check_tau,
    LogImprovementMCAcquisitionFunction,
)
from botorch.acquisition.objective import MCAcquisitionObjective, PosteriorTransform
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.utils.safe_math import logmeanexp
from botorch.utils.transforms import (
    average_over_ensemble_models,
    t_batch_mode_transform,
)
from torch import Tensor

TAU_RELU = 1e-6
TAU_MAX = 1e-2


class LogRegionalExpectedImprovement(AnalyticAcquisitionFunction):
    _log: bool = True

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        X_dev: Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
        length: float = 0.8,
        bounds: Tensor | None = None,
    ) -> None:
        r"""Log-Regional Expected Improvement (analytic).

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            X_dev: A `n x d`-dim Tensor of `n` `d`-dim design points within a TR.
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
            length: The length of the trust region to consider.
            bounds: The bounds of the design space.
                First column represents dimension-wise lower bounds.
                Second column represents dimension-wise upper bounds.
        """

        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.maximize: bool = maximize

        dim: int = X_dev.shape[1]
        self.n_region: int = X_dev.shape[0]
        self.X_dev: Tensor = X_dev.reshape(self.n_region, 1, 1, -1)
        self.length: float = length
        if bounds is not None:
            self.bounds = bounds
        else:
            self.bounds = torch.stack([torch.zeros(dim), torch.ones(dim)]).to(
                device=self.X_dev.device, dtype=self.X_dev.dtype
            )

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        batch_shape = X.shape[0]
        q = X.shape[1]
        d = X.shape[2]

        # make N_x samples in design space
        X_min = (X - 0.5 * self.length).clamp_min(self.bounds[0]).unsqueeze(0)
        X_max = (X + 0.5 * self.length).clamp_max(self.bounds[1]).unsqueeze(0)
        Xs = (self.X_dev * (X_max - X_min) + X_min).reshape(-1, q, d)

        mean, sigma = self._mean_and_sigma(Xs)
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        logei = _log_ei_helper(u) + sigma.log()
        logrei = logmeanexp(logei.reshape(self.n_region, batch_shape), dim=0)
        return logrei


class qLogRegionalExpectedImprovement(LogImprovementMCAcquisitionFunction):
    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        X_dev: Tensor,
        sampler: MCSampler | None = None,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
        X_pending: Tensor | None = None,
        length: float = 0.8,
        bounds: Tensor | None = None,
        fat: bool = True,
        tau_relu: float = TAU_RELU,
    ) -> None:
        r"""q-Log Regional Expected Improvement (MC acquisition function).

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            X_dev: A `n x d`-dim Tensor of `n` `d`-dim design points within a TR.
            sampler: botorch.sampling.base.MCSampler
                The sampler used to sample fantasized models. Defaults to
                SobolQMCNormalSampler(num_samples=1)`.
            objective: The MCAcquisitionObjective under which the samples are evaluated.
                Defaults to `IdentityMCObjective()`.
                NOTE: `ConstrainedMCObjective` for outcome constraints is deprecated in
                favor of passing the `constraints` directly to this constructor.
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` `d`-dim design
                points that have been submitted for function evaluation but have
                not yet been evaluated.
            length: The length of the trust region to consider.
            bounds: The bounds of the design space.
                First column represents dimension-wise lower bounds.
                Second column represents dimension-wise upper bounds.
            fat: Toggles the logarithmic / linear asymptotic behavior of the smooth
                approximation to the ReLU.
            tau_relu: Temperature parameter controlling the sharpness of the smooth
                approximations to ReLU.

        """
        super().__init__(
            model=model,
            sampler=sampler,
            objective=objective,
            posterior_transform=posterior_transform,
            X_pending=X_pending,
        )
        # adding + 1 to account for the additional MC sampling dimension
        # for points inside the trust region surrounding `X`
        sample_dim = tuple(range(len(self.sample_shape) + 1))
        self._sample_reduction = partial(logmeanexp, dim=sample_dim)

        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.fat: bool = fat
        self.tau_relu: float = check_tau(tau_relu, "tau_relu")
        dim: int = X_dev.shape[1]
        self.n_region: int = X_dev.shape[0]
        self.X_dev: Tensor = X_dev
        self.length: float = length
        if bounds is not None:
            self.bounds = bounds
        else:
            self.bounds = torch.stack([torch.zeros(dim), torch.ones(dim)]).to(
                device=self.X_dev.device, dtype=self.X_dev.dtype
            )

    def _get_samples_and_objectives(self, X: Tensor) -> tuple[Tensor, Tensor]:
        """Computes posterior samples and objective values at input X.

        This is required in order to munge X and create samples within a TR
        and compute the MC objective over these points. This is then later
        passed to the _sample_forward function to compute the Log Regional
        Expected Improvement acqusition function value.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of model inputs.

        Returns:
            A two-tuple `(samples, obj)`, where `samples` is a tensor of posterior
            samples with shape `sample_shape x n_region x batch_shape x q x m`,
            and `obj` is a tensor of MC objective values with shape
            `sample_shape x n_region x batch_shape x q`.
        """
        # region-averaged EI specific code
        batch_shape = X.shape[:-2]
        d = X.shape[2]

        # make N_x samples in design space
        X_min = (X - 0.5 * self.length).clamp_min(self.bounds[0]).unsqueeze(0)
        X_max = (X + 0.5 * self.length).clamp_max(self.bounds[1]).unsqueeze(0)

        # n_region x (1, ..., 1) x 1 x d
        X_dev = self.X_dev.reshape(
            (self.n_region,) + (tuple(1 for _ in batch_shape) + (1, d))
        )
        # n_region x batch_shape x q x d
        Xs = X_dev * (X_max - X_min) + X_min

        # calling the original method with the modified inputs
        return super()._get_samples_and_objectives(Xs)

    def _sample_forward(self, obj: Tensor) -> Tensor:
        r"""Evaluate qLogExpectedImprovement on the candidate set `X`.

        Args:
            obj: `mc_shape x batch_shape x q`-dim Tensor of MC objective values.

        Returns:
            A `mc_shape x batch_shape x q`-dim Tensor of expected improvement values.
        """
        li = _log_improvement(
            Y=obj,
            best_f=self.best_f,
            tau=self.tau_relu,
            fat=self._fat,
        )
        return li



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import os
from logging import DEBUG, WARN
from unittest.mock import MagicMock, mock_open, patch

import torch
from botorch.acquisition.objective import ScalarizedPosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.models.transforms.input import Normalize
from botorch.utils.testing import BotorchTestCase
from botorch_community.models.prior_fitted_network import (
    BoundedRiemannPosterior,
    MultivariatePFNModel,
    PFNModel,
)
from botorch_community.models.utils.prior_fitted_network import (
    download_model,
    ModelPaths,
)
from botorch_community.posteriors.riemann import MultivariateRiemannPosterior
from pfns.model.transformer_config import CrossEntropyConfig, TransformerConfig
from pfns.train import MainConfig, OptimizerConfig
from torch import nn, Tensor


class DummyPFN(nn.Module):
    def __init__(self, n_buckets: int = 1000):
        """A dummy PFN model for testing purposes.

        This class implements a mocked PFN model that returns
        constant values for testing. It mimics the interface of actual PFN models
        but with simplified behavior.

        Args:
            n_buckets: Number of buckets for the output distribution. Default is 1000.
        """

        super().__init__()
        self.n_buckets = n_buckets
        self.criterion = MagicMock()
        self.criterion.borders = torch.linspace(0, 1, n_buckets + 1)

    def forward(self, train_X: Tensor, train_Y: Tensor, test_X: Tensor) -> Tensor:
        return torch.zeros(*test_X.shape[:-1], self.n_buckets, device=test_X.device)


class TestPriorFittedNetwork(BotorchTestCase):
    def test_raises(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            train_X = torch.rand(10, 3, **tkwargs)
            train_Y = torch.rand(10, 1, **tkwargs)
            train_Yvar = torch.rand(10, 1, **tkwargs)
            test_X = torch.rand(5, 3, **tkwargs)

            with self.assertLogs(logger="botorch", level=DEBUG) as log:
                PFNModel(train_X, train_Y, DummyPFN(), train_Yvar=train_Yvar)
                self.assertIn(
                    "train_Yvar provided but ignored for PFNModel.",
                    log.output[0],
                )

            train_Y_4d = torch.rand(10, 2, 2, 1, **tkwargs)
            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, train_Y_4d, DummyPFN())

            train_Y_2d = torch.rand(10, 2, **tkwargs)
            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, train_Y_2d, DummyPFN())

            with self.assertRaises(UnsupportedError):
                PFNModel(torch.rand(10, 3, 3, 2, **tkwargs), train_Y, DummyPFN())

            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, torch.rand(11, **tkwargs), DummyPFN())

            pfn = PFNModel(train_X, train_Y, DummyPFN())

            with self.assertRaises(UnsupportedError):
                pfn.posterior(test_X, output_indices=[0, 1])
            with self.assertLogs(logger="botorch", level=WARN) as log:
                pfn.posterior(test_X, observation_noise=True)
                self.assertIn(
                    "observation_noise is not supported for PFNModel",
                    log.output[0],
                )
            with self.assertRaises(UnsupportedError):
                pfn.posterior(
                    test_X,
                    posterior_transform=ScalarizedPosteriorTransform(
                        weights=torch.ones(1)
                    ),
                )

            # (b', b, d) prediction works as expected
            test_X = torch.rand(5, 4, 2, **tkwargs)
            post = pfn.posterior(test_X)
            self.assertEqual(post.mean.shape, torch.Size([5, 4, 1]))

            # X dims should be 1 to 4
            test_X = torch.rand(5, 4, 2, 1, 2, **tkwargs)
            with self.assertRaises(UnsupportedError):
                pfn.posterior(test_X)

    def test_shapes(self):
        tkwargs = {"device": self.device, "dtype": torch.float}

        # no q dimension
        train_X = torch.rand(10, 3, **tkwargs)
        train_Y = torch.rand(10, 1, **tkwargs)
        test_X = torch.rand(5, 3, **tkwargs)

        pfn = PFNModel(train_X, train_Y, DummyPFN(n_buckets=100))

        for batch_first in [True, False]:
            with self.subTest(batch_first=batch_first):
                pfn.batch_first = batch_first
                posterior = pfn.posterior(test_X)

                self.assertEqual(posterior.probabilities.shape, torch.Size([5, 100]))
                self.assertEqual(posterior.mean.shape, torch.Size([5, 1]))

        # q=1
        test_X = torch.rand(5, 1, 3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([5, 1, 100]))
        self.assertEqual(posterior.mean.shape, torch.Size([5, 1, 1]))

        # no shape basically
        test_X = torch.rand(3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([100]))
        self.assertEqual(posterior.mean.shape, torch.Size([1]))

        # prepare_data
        X = torch.rand(5, 3, **tkwargs)
        X, train_X, train_Y, orig_X_shape = pfn._prepare_data(X)
        self.assertEqual(X.shape, torch.Size([1, 5, 3]))
        self.assertEqual(train_X.shape, torch.Size([1, 10, 3]))
        self.assertEqual(train_Y.shape, torch.Size([1, 10, 1]))
        self.assertEqual(orig_X_shape, torch.Size([5, 3]))

    def test_input_transform(self):
        model = PFNModel(
            train_X=torch.rand(10, 3),
            train_Y=torch.rand(10, 1),
            input_transform=Normalize(d=3),
            model=DummyPFN(),
        )
        self.assertIsInstance(model.input_transform, Normalize)
        self.assertEqual(model.input_transform.bounds.shape, torch.Size([2, 3]))

    def test_unpack_checkpoint(self):
        config = MainConfig(
            priors=[],
            optimizer=OptimizerConfig(
                optimizer="adam",
                lr=0.001,
            ),
            model=TransformerConfig(
                criterion=CrossEntropyConfig(num_classes=3),
            ),
            batch_shape_sampler=None,
        )

        model = config.model.create_model()

        state_dict = model.state_dict()
        checkpoint = {
            "config": config.to_dict(),
            "model_state_dict": state_dict,
        }

        loaded_model = PFNModel(
            train_X=torch.rand(10, 3),
            train_Y=torch.rand(10, 1),
            input_transform=Normalize(d=3),
            model=checkpoint,
            load_training_checkpoint=True,
        )

        loaded_state_dict = loaded_model.pfn.state_dict()
        self.assertEqual(
            sorted(loaded_state_dict.keys()),
            sorted(state_dict.keys()),
        )
        for k in loaded_state_dict.keys():
            self.assertTrue(torch.equal(loaded_state_dict[k], state_dict[k]))


class TestPriorFittedNetworkUtils(BotorchTestCase):
    @patch("botorch_community.models.utils.prior_fitted_network.requests.get")
    @patch("botorch_community.models.utils.prior_fitted_network.gzip.GzipFile")
    @patch("botorch_community.models.utils.prior_fitted_network.torch.load")
    @patch("botorch_community.models.utils.prior_fitted_network.torch.save")
    @patch("botorch_community.models.utils.prior_fitted_network.os.path.exists")
    @patch("botorch_community.models.utils.prior_fitted_network.os.makedirs")
    def test_download_model_cache_miss(
        self,
        _mock_makedirs,
        mock_exists,
        mock_torch_save,
        mock_torch_load,
        mock_gzip,
        mock_requests_get,
    ):
        # Simulate cache miss
        mock_exists.return_value = False

        # Mock the requests.get to simulate a network call
        mock_requests_get.return_value = MagicMock(
            status_code=200, content=b"fake content"
        )

        # Mock the gzip.GzipFile to simulate decompression
        mock_gzip.return_value.__enter__.return_value = mock_open(
            read_data=b"fake model data"
        ).return_value

        # Mock torch.load to simulate loading a model
        fake_model = MagicMock(spec=torch.nn.Module)
        mock_torch_load.return_value = fake_model

        # Call the function
        model = download_model(
            ModelPaths.pfns4bo_hebo,
            cache_dir=os.environ.get("RUNNER_TEMP", "/tmp") + "/test_cache",
            # $RUNNER_TEMP is set by GitHub Actions as tmp, /tmp does not work there
        )

        # Assertions for cache miss
        mock_requests_get.assert_called_once()
        mock_gzip.assert_called_once()
        mock_torch_load.assert_called_once()
        mock_torch_save.assert_called_once()
        self.assertEqual(model, fake_model)

        # Test loading in model init
        model = PFNModel(
            train_X=torch.rand(10, 3),
            train_Y=torch.rand(10, 1),
        )
        self.assertEqual(model.pfn, fake_model.to("cpu"))

    @patch("botorch_community.models.utils.prior_fitted_network.torch.load")
    @patch("botorch_community.models.utils.prior_fitted_network.os.path.exists")
    def test_download_model_cache_hit(self, mock_exists, mock_torch_load):
        # Simulate cache hit
        mock_exists.return_value = True

        # Mock torch.load to simulate loading a model
        fake_model = MagicMock(spec=torch.nn.Module)
        mock_torch_load.return_value = fake_model

        # Call the function
        model = download_model(
            ModelPaths.pfns4bo_hebo,
            cache_dir=os.environ.get("RUNNER_TEMP", "/tmp") + "/test_cache",
            # $RUNNER_TEMP is set by GitHub Actions as tmp, /tmp does not work there
        )

        # Assertions for cache hit
        # mock_exists is called once here and once through os.makedirs
        # which checks if directory exists
        self.assertEqual(mock_exists.call_count, 2)
        mock_torch_load.assert_called_once()
        self.assertEqual(model, fake_model)


class TestMultivariatePFN(BotorchTestCase):
    def setUp(self):
        train_X = torch.rand(10, 3)
        train_Y = torch.rand(10, 1)
        self.pfn = MultivariatePFNModel(train_X, train_Y, DummyPFN())

    def test_posterior(self):
        X = torch.rand(1, 3)
        post = self.pfn.posterior(X)
        self.assertNotIsInstance(post, MultivariateRiemannPosterior)
        X = torch.rand(4, 3)
        R = torch.rand(1, 4, 4)
        with patch(
            "botorch_community.models.prior_fitted_network.MultivariatePFNModel"
            ".estimate_correlations",
            return_value=R,
        ):
            post = self.pfn.posterior(X)
        self.assertIsInstance(post, MultivariateRiemannPosterior)
        self.assertTrue(torch.equal(post.correlation_matrix, R.squeeze(0)))

    def test_estimate_covariances(self):
        b = 3
        q = 4
        cond_val = torch.rand(b, q)
        cond_mean = torch.rand(b, q, q)
        var = torch.ones(b, q)
        mean = torch.rand(b, q)
        # Fill in particular values for the [1, 1, 2] entries
        mean[1, 1] = 2.0
        mean[1, 2] = 3.0
        cond_mean[1, 2, 1] = 3.0
        cond_mean[1, 1, 2] = 4.0
        cond_mean[1, 1, 1] = 2.1
        cond_mean[1, 2, 2] = 3.1
        cond_val[1, 1] = 3.0
        cond_val[1, 2] = 4.0
        with patch(
            "botorch_community.models.prior_fitted_network.MultivariatePFNModel."
            "_map_psd"
        ) as mock_map_psd:
            self.pfn._estimate_covariances(
                cond_mean=cond_mean, cond_val=cond_val, mean=mean, var=var
            )
        cov = mock_map_psd.call_args[0][0]
        # Compare to analytical value of 10
        self.assertEqual(torch.round(cov[1, 1, 2], decimals=2).item(), 10.0)
        self.assertEqual(torch.round(cov[1, 2, 1], decimals=2).item(), 10.0)

    def test_compute_conditional_means(self):
        probabilities = torch.zeros(3, 2, 1000)
        probabilities[0, 0, 9] = 1.0
        probabilities[0, 1, 19] = 1.0
        probabilities[1, 0, 29] = 1.0
        probabilities[1, 1, 39] = 1.0
        probabilities[2, 0, 49] = 1.0
        probabilities[2, 1, 59] = 1.0
        marginals = BoundedRiemannPosterior(
            borders=self.pfn.borders,
            probabilities=probabilities,
        )
        return_value = torch.zeros(3 * 2, 2, 1000)
        return_value[..., 100] = 1.0
        X = torch.ones(3, 2, 5)
        X[:, 1, :] = 2.0
        with patch(
            "botorch_community.models.prior_fitted_network.MultivariatePFNModel."
            "pfn_predict",
            return_value=return_value,
        ) as mock_pfn_predict:
            self.pfn._compute_conditional_means(
                X=X,
                train_X=torch.zeros(3, 4, 5),
                train_Y=torch.zeros(3, 4, 1),
                marginals=marginals,
            )
        res = mock_pfn_predict.call_args[1]
        self.assertTrue(torch.equal(res["X"], torch.cat([X, X])))
        X1 = torch.zeros(1, 5, 5)
        X1[:, -1, :] = 1.0
        X2 = torch.zeros(1, 5, 5)
        X2[:, -1, :] = 2.0
        self.assertTrue(
            torch.equal(res["train_X"], torch.cat([X1, X2, X1, X2, X1, X2], dim=0))
        )
        a = []
        for i in range(6):
            Y = torch.zeros(1, 5, 1)
            Y[0, -1, 0] = (i + 1) * 0.01
            a.append(Y)
        self.assertTrue(
            torch.equal(torch.round(res["train_Y"], decimals=2), torch.cat(a, dim=0))
        )

    def test_estimate_correlations(self):
        probabilities = torch.ones(2, 3, 1000)
        probabilities = probabilities / probabilities.sum(dim=-1, keepdim=True)
        marginals = BoundedRiemannPosterior(
            borders=self.pfn.borders,
            probabilities=probabilities,
        )
        cond_mean = 0.5 * (1 + torch.rand(2, 3, 3))
        with patch(
            "botorch_community.models.prior_fitted_network.MultivariatePFNModel."
            "_compute_conditional_means",
            return_value=(cond_mean, 0.9 * torch.ones(2, 3)),
        ):
            R = self.pfn.estimate_correlations(
                X=torch.ones(2, 3, 5),
                train_X=torch.zeros(2, 4, 5),
                train_Y=torch.zeros(2, 4, 1),
                marginals=marginals,
            )
        self.assertAllClose(torch.diagonal(R, dim1=-2, dim2=-1), torch.ones(2, 3))
        # Test with no batch dimension
        marginals = BoundedRiemannPosterior(
            borders=self.pfn.borders,
            probabilities=probabilities[0, ...],
        )
        cond_mean = cond_mean[:1, ...]
        with patch(
            "botorch_community.models.prior_fitted_network.MultivariatePFNModel."
            "_compute_conditional_means",
            return_value=(cond_mean, 0.9 * torch.ones(1, 3)),
        ):
            R = self.pfn.estimate_correlations(
                X=torch.ones(1, 3, 5),
                train_X=torch.zeros(1, 4, 5),
                train_Y=torch.zeros(1, 4, 1),
                marginals=marginals,
            )
        self.assertEqual(R.shape, torch.Size([1, 3, 3]))
        self.assertAllClose(torch.diagonal(R, dim1=-2, dim2=-1), torch.ones(1, 3))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Sampler to be used with `EnsemblePosteriors` to enable
deterministic optimization of acquisition functions with ensemble models.
"""

from __future__ import annotations

import torch
from botorch.posteriors.ensemble import EnsemblePosterior
from botorch.sampling.base import MCSampler
from torch import Tensor
from torch.distributions.multinomial import Multinomial


class IndexSampler(MCSampler):
    r"""A sampler that calls `posterior.rsample_from_base_samples` to
    generate the samples via index base samples."""

    def forward(self, posterior: EnsemblePosterior) -> Tensor:
        r"""Draws MC samples from the posterior.

        Args:
            posterior: The ensemble posterior to sample from.

        Returns:
            The samples drawn from the posterior.
        """
        self._construct_base_samples(posterior=posterior)
        samples = posterior.rsample_from_base_samples(
            sample_shape=self.sample_shape, base_samples=self.base_samples
        )
        return samples

    def _construct_base_samples(self, posterior: EnsemblePosterior) -> None:
        r"""Constructs base samples as indices to sample with them from
        the Posterior.

        Args:
            posterior: The ensemble posterior to construct the base samples
                for.
        """
        if (
            self.base_samples is None
            or self.base_samples.shape != self.sample_shape + posterior.batch_shape
        ):
            with torch.random.fork_rng():
                torch.manual_seed(self.seed)
                base_samples = (
                    Multinomial(
                        probs=posterior.mixture_weights,
                    )
                    .sample(sample_shape=self.sample_shape)
                    .argmax(dim=-1)
                )
            self.register_buffer("base_samples", base_samples)
        if self.base_samples.device != posterior.device:
            self.to(device=posterior.device)  # pragma: nocover

    def _update_base_samples(
        self, posterior: EnsemblePosterior, base_sampler: IndexSampler
    ) -> None:
        r"""Null operation just needed for compatibility with
        `CachedCholeskyAcquisitionFunction`."""
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


r"""Multi-task Gaussian Process Regression models with fully Bayesian inference."""

from collections.abc import Mapping
from typing import Any, NoReturn, TypeVar

import pyro
import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.models.fully_bayesian import (
    matern52_kernel,
    MCMC_DIM,
    MIN_INFERRED_NOISE_LEVEL,
    reshape_and_detach,
    SaasPyroModel,
)
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel
from botorch.models.multitask import MultiTaskGP
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.posteriors.fully_bayesian import GaussianMixturePosterior
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import MaternKernel
from gpytorch.kernels.index_kernel import IndexKernel
from gpytorch.kernels.kernel import Kernel
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.means.mean import Mean
from torch import Tensor
from torch.nn.parameter import Parameter

# Can replace with Self type once 3.11 is the minimum version
TSaasFullyBayesianMultiTaskGP = TypeVar(
    "TSaasFullyBayesianMultiTaskGP", bound="SaasFullyBayesianMultiTaskGP"
)


class MultitaskSaasPyroModel(SaasPyroModel):
    r"""
    Implementation of the multi-task sparse axis-aligned subspace priors (SAAS) model.

    The multi-task model uses an ICM kernel. The data kernel is same as the single task
    SAAS model in order to handle high-dimensional parameter spaces. The task kernel
    is a Matern-5/2 kernel using learned task embeddings as the input.
    """

    def set_inputs(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Tensor | None,
        task_feature: int,
        task_rank: int | None = None,
    ) -> None:
        """Set the training data.

        Args:
            train_X: Training inputs (n x (d + 1))
            train_Y: Training targets (n x 1)
            train_Yvar: Observed noise variance (n x 1). If None, we infer the noise.
                Note that the inferred noise is common across all tasks.
            task_feature: The index of the task feature (`-d <= task_feature <= d`).
            task_rank: The num of learned task embeddings to be used in the task kernel.
                If omitted, use a full rank (i.e. number of tasks) kernel.
        """
        # NOTE PyTorch does not support negative indexing for tensors in index_select,
        # (https://github.com/pytorch/pytorch/issues/76347), so we have to make sure
        # that the task feature is positive.
        task_feature = task_feature % train_X.shape[-1]
        super().set_inputs(train_X, train_Y, train_Yvar)
        # obtain a list of task indicies
        all_tasks = train_X[:, task_feature].unique().to(dtype=torch.long).tolist()
        self.task_feature = task_feature
        self.num_tasks = len(all_tasks)
        self.task_rank = task_rank or self.num_tasks
        # assume there is one column for task feature
        self.ard_num_dims = self.train_X.shape[-1] - 1

    def sample(self) -> None:
        r"""Sample from the SAAS model.

        This samples the mean, noise variance, outputscale, and lengthscales according
        to the SAAS prior.
        """
        tkwargs = {"dtype": self.train_X.dtype, "device": self.train_X.device}
        base_idxr = torch.arange(self.ard_num_dims, **{"device": tkwargs["device"]})
        base_idxr[self.task_feature :] += 1  # exclude task feature
        task_indices = self.train_X[..., self.task_feature].to(
            device=tkwargs["device"], dtype=torch.long
        )

        outputscale = self.sample_outputscale(concentration=2.0, rate=0.15, **tkwargs)
        mean = self.sample_mean(**tkwargs)
        noise = self.sample_noise(**tkwargs)

        lengthscale = self.sample_lengthscale(dim=self.ard_num_dims, **tkwargs)
        K = matern52_kernel(X=self.train_X[..., base_idxr], lengthscale=lengthscale)

        # compute task covar matrix
        task_latent_features = self.sample_latent_features(**tkwargs)[task_indices]
        task_lengthscale = self.sample_task_lengthscale(**tkwargs)
        task_covar = matern52_kernel(
            X=task_latent_features, lengthscale=task_lengthscale
        )
        K = K.mul(task_covar)
        K = outputscale * K + noise * torch.eye(self.train_X.shape[0], **tkwargs)
        pyro.sample(
            "Y",
            pyro.distributions.MultivariateNormal(
                loc=mean.view(-1).expand(self.train_X.shape[0]),
                covariance_matrix=K,
            ),
            obs=self.train_Y.squeeze(-1),
        )

    def sample_latent_features(self, **tkwargs: Any):
        return pyro.sample(
            "latent_features",
            pyro.distributions.Normal(
                torch.tensor(0.0, **tkwargs),
                torch.tensor(1.0, **tkwargs),
            ).expand(torch.Size([self.num_tasks, self.task_rank])),
        )

    def sample_task_lengthscale(
        self, concentration: float = 6.0, rate: float = 3.0, **tkwargs: Any
    ):
        return pyro.sample(
            "task_lengthscale",
            pyro.distributions.Gamma(
                torch.tensor(concentration, **tkwargs),
                torch.tensor(rate, **tkwargs),
            ).expand(torch.Size([self.task_rank])),
        )

    def load_mcmc_samples(
        self, mcmc_samples: dict[str, Tensor]
    ) -> tuple[Mean, Kernel, Likelihood, Kernel]:
        r"""Load the MCMC samples into the mean_module, covar_module, and likelihood."""
        tkwargs = {"device": self.train_X.device, "dtype": self.train_X.dtype}
        num_mcmc_samples = len(mcmc_samples["mean"])
        batch_shape = torch.Size([num_mcmc_samples])

        mean_module, data_covar_module, likelihood, _ = super().load_mcmc_samples(
            mcmc_samples=mcmc_samples
        )
        data_indices = torch.arange(self.train_X.shape[-1] - 1)
        data_indices[self.task_feature :] += 1  # exclude task feature

        data_covar_module.active_dims = data_indices.to(device=tkwargs["device"])
        latent_covar_module = MaternKernel(
            nu=2.5,
            ard_num_dims=self.task_rank,
            batch_shape=batch_shape,
        ).to(**tkwargs)

        latent_covar_module.lengthscale = reshape_and_detach(
            target=latent_covar_module.lengthscale,
            new_value=mcmc_samples["task_lengthscale"],
        )
        latent_features = mcmc_samples["latent_features"]
        task_covar = latent_covar_module(latent_features)
        task_covar_module = IndexKernel(
            num_tasks=self.num_tasks,
            rank=self.task_rank,
            batch_shape=latent_features.shape[:-2],
            active_dims=torch.tensor([self.task_feature], device=tkwargs["device"]),
        )
        task_covar_module.covar_factor = Parameter(
            task_covar.cholesky().to_dense().detach()
        )
        task_covar_module = task_covar_module.to(**tkwargs)
        # NOTE: The IndexKernel has a learnable 'var' parameter in addition to the
        # task covariances, corresponding do task-specific variances along the diagonal
        # of the task covariance matrix. As this parameter is not sampled in `sample()`
        # we implicitly assume it to be zero. This is consistent with the previous
        # SAASFBMTGP implementation, but means that the non-fully Bayesian and fully
        # Bayesian models run on slightly different task covar modules.

        # We set the aforementioned task covar module var parameter to zero here.
        task_covar_module.var = torch.zeros_like(task_covar_module.var)
        covar_module = data_covar_module * task_covar_module
        return mean_module, covar_module, likelihood, None


class SaasFullyBayesianMultiTaskGP(MultiTaskGP):
    r"""A fully Bayesian multi-task GP model with the SAAS prior.
    This model assumes that the inputs have been normalized to [0, 1]^d and that the
    output has been stratified standardized to have zero mean and unit variance for
    each task. The SAAS model [Eriksson2021saasbo]_ with a Matern-5/2 is used as data
    kernel by default.

    You are expected to use `fit_fully_bayesian_model_nuts` to fit this model as it
    isn't compatible with `fit_gpytorch_mll`.

    Example:
        >>> X1, X2 = torch.rand(10, 2), torch.rand(20, 2)
        >>> i1, i2 = torch.zeros(10, 1), torch.ones(20, 1)
        >>> train_X = torch.cat([
        >>>     torch.cat([X1, i1], -1), torch.cat([X2, i2], -1),
        >>> ])
        >>> train_Y = torch.cat(f1(X1), f2(X2)).unsqueeze(-1)
        >>> train_Yvar = 0.01 * torch.ones_like(train_Y)
        >>> mtsaas_gp = SaasFullyBayesianMultiTaskGP(
        >>>     train_X, train_Y, train_Yvar, task_feature=-1,
        >>> )
        >>> fit_fully_bayesian_model_nuts(mtsaas_gp)
        >>> posterior = mtsaas_gp.posterior(test_X)
    """

    _is_fully_bayesian = True
    _is_ensemble = True

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        task_feature: int,
        train_Yvar: Tensor | None = None,
        output_tasks: list[int] | None = None,
        rank: int | None = None,
        all_tasks: list[int] | None = None,
        outcome_transform: OutcomeTransform | None = None,
        input_transform: InputTransform | None = None,
        pyro_model: MultitaskSaasPyroModel | None = None,
        validate_task_values: bool = True,
    ) -> None:
        r"""Initialize the fully Bayesian multi-task GP model.

        Args:
            train_X: Training inputs (n x (d + 1))
            train_Y: Training targets (n x 1)
            train_Yvar: Observed noise variance (n x 1). If None, we infer the noise.
                Note that the inferred noise is common across all tasks.
            task_feature: The index of the task feature (`-d <= task_feature <= d`).
            output_tasks: A list of task indices for which to compute model
                outputs for. If omitted, return outputs for all task indices.
            rank: The num of learned task embeddings to be used in the task kernel.
                If omitted, use a full rank (i.e. number of tasks) kernel.
            all_tasks: NOT SUPPORTED!
            outcome_transform: An outcome transform that is applied to the
                training data during instantiation and to the posterior during
                inference (that is, the `Posterior` obtained by calling
                `.posterior` on the model will be on the original scale).
                Note that `.train()` will be called on the outcome transform during
                instantiation of the model.
            input_transform: An input transform that is applied to the inputs `X`
                in the model's forward pass.
            pyro_model: Optional `PyroModel` that has the same signature as
                `MultitaskSaasPyroModel`. Defaults to `MultitaskSaasPyroModel`.
            validate_task_values: If True, validate that the task values supplied in the
                input are expected tasks values. If false, unexpected task values
                will be mapped to the first output_task if supplied.
        """
        if not (
            train_X.ndim == train_Y.ndim == 2
            and len(train_X) == len(train_Y)
            and train_Y.shape[-1] == 1
        ):
            raise ValueError(
                "Expected train_X to have shape n x d and train_Y to have shape n x 1"
            )
        if train_Yvar is not None and train_Y.shape != train_Yvar.shape:
            raise ValueError(
                "Expected train_Yvar to be None or have the same shape as train_Y"
            )
        with torch.no_grad():
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )
        if outcome_transform is not None:
            outcome_transform.train()  # Ensure we learn parameters here on init
            train_Y, train_Yvar = outcome_transform(
                Y=train_Y, Yvar=train_Yvar, X=transformed_X
            )
        if train_Yvar is not None:  # Clamp after transforming
            train_Yvar = train_Yvar.clamp(MIN_INFERRED_NOISE_LEVEL)

        super().__init__(
            train_X=train_X,
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            task_feature=task_feature,
            output_tasks=output_tasks,
            rank=rank,
            # We already transformed the data above, this avoids applying the
            # default `Standardize` transform twice. As outcome_transform is
            # set on `self` below, it will be applied to the posterior in the
            # `posterior` method of `MultiTaskGP`.
            outcome_transform=None,
            all_tasks=all_tasks,
            validate_task_values=validate_task_values,
        )
        self.to(train_X)
        self.mean_module = None
        self.covar_module = None
        self.likelihood = None
        if pyro_model is None:
            pyro_model = MultitaskSaasPyroModel()
        # apply task_mapper
        x_before, task_idcs, x_after = self._split_inputs(transformed_X)
        pyro_model.set_inputs(
            train_X=torch.cat([x_before, task_idcs, x_after], dim=-1),
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            task_feature=task_feature,
            task_rank=self._rank,
        )
        self.pyro_model: MultitaskSaasPyroModel = pyro_model
        if outcome_transform is not None:
            self.outcome_transform = outcome_transform
        if input_transform is not None:
            self.input_transform = input_transform

    def train(
        self, mode: bool = True, reset: bool = True
    ) -> TSaasFullyBayesianMultiTaskGP:
        r"""Puts the model in `train` mode.

        Args:
            mode: A boolean indicating whether to put the model in training mode.
            reset: A boolean indicating whether to reset the model to its initial
                state. If `mode` is False, this argument is ignored.

        Returns:
            The model itself.
        """
        super().train(mode=mode)
        if mode and reset:
            self.mean_module = None
            self.covar_module = None
            self.likelihood = None
        return self

    @property
    def median_lengthscale(self) -> Tensor:
        r"""Median lengthscales across the MCMC samples."""
        self._check_if_fitted()
        lengthscale = self.covar_module.kernels[0].base_kernel.lengthscale.clone()
        return lengthscale.median(0).values.squeeze(0)

    @property
    def num_mcmc_samples(self) -> int:
        r"""Number of MCMC samples in the model."""
        self._check_if_fitted()
        return self.covar_module.kernels[0].batch_shape[0]

    @property
    def batch_shape(self) -> torch.Size:
        r"""Batch shape of the model, equal to the number of MCMC samples.
        Note that `SaasFullyBayesianMultiTaskGP` does not support batching
        over input data at this point.
        """
        self._check_if_fitted()
        return torch.Size([self.num_mcmc_samples])

    def fantasize(self, *args, **kwargs) -> NoReturn:
        raise NotImplementedError("Fantasize is not implemented!")

    def _check_if_fitted(self):
        r"""Raise an exception if the model hasn't been fitted."""
        if self.covar_module is None:
            raise RuntimeError(
                "Model has not been fitted. You need to call "
                "`fit_fully_bayesian_model_nuts` to fit the model."
            )

    def load_mcmc_samples(self, mcmc_samples: dict[str, Tensor]) -> None:
        r"""Load the MCMC hyperparameter samples into the model.

        This method will be called by `fit_fully_bayesian_model_nuts` when the model
        has been fitted in order to create a batched MultiTaskGP model.
        """
        (
            self.mean_module,
            self.covar_module,
            self.likelihood,
            _,
        ) = self.pyro_model.load_mcmc_samples(mcmc_samples=mcmc_samples)

    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool = False,
        posterior_transform: PosteriorTransform | None = None,
        **kwargs: Any,
    ) -> GaussianMixturePosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Returns:
            A `GaussianMixturePosterior` object. Includes observation noise
                if specified.
        """
        self._check_if_fitted()
        posterior = super().posterior(
            X=X.unsqueeze(MCMC_DIM),
            output_indices=output_indices,
            observation_noise=observation_noise,
            posterior_transform=posterior_transform,
            **kwargs,
        )
        posterior = GaussianMixturePosterior(distribution=posterior.distribution)
        return posterior

    def forward(self, X: Tensor) -> MultivariateNormal:
        self._check_if_fitted()
        return super().forward(X)

    def load_state_dict(self, state_dict: Mapping[str, Any], strict: bool = True):
        r"""Custom logic for loading the state dict.

        The standard approach of calling `load_state_dict` currently doesn't play well
        with the `SaasFullyBayesianMultiTaskGP` since the `mean_module`, `covar_module`
        and `likelihood` aren't initialized until the model has been fitted. The reason
        for this is that we don't know the number of MCMC samples until NUTS is called.
        Given the state dict, we can initialize a new model with some dummy samples and
        then load the state dict into this model. This currently only works for a
        `MultitaskSaasPyroModel` and supporting more Pyro models likely requires moving
        the model construction logic into the Pyro model itself.

        TODO: If this were to inherif from `SaasFullyBayesianSingleTaskGP`, we could
        simplify this method and eliminate some others.
        """
        if not isinstance(self.pyro_model, MultitaskSaasPyroModel):
            raise NotImplementedError(  # pragma: no cover
                "load_state_dict only works for MultitaskSaasPyroModel"
            )
        raw_mean = state_dict["mean_module.raw_constant"]
        num_mcmc_samples = len(raw_mean)
        dim = self.pyro_model.train_X.shape[-1] - 1  # Removing 1 for the task feature.
        tkwargs = {"device": raw_mean.device, "dtype": raw_mean.dtype}
        # Load some dummy samples
        mcmc_samples = {
            "mean": torch.ones(num_mcmc_samples, **tkwargs),
            "lengthscale": torch.ones(num_mcmc_samples, dim, **tkwargs),
            "outputscale": torch.ones(num_mcmc_samples, **tkwargs),
            "task_lengthscale": torch.ones(num_mcmc_samples, self._rank, **tkwargs),
            "latent_features": torch.ones(
                num_mcmc_samples, self.num_tasks, self._rank, **tkwargs
            ),
        }
        if self.pyro_model.train_Yvar is None:
            mcmc_samples["noise"] = torch.ones(num_mcmc_samples, **tkwargs)
        (
            self.mean_module,
            self.covar_module,
            self.likelihood,
            _,  # Possibly space for input transform
        ) = self.pyro_model.load_mcmc_samples(mcmc_samples=mcmc_samples)
        # Load the actual samples from the state dict
        super().load_state_dict(state_dict=state_dict, strict=strict)

    def condition_on_observations(
        self, X: Tensor, Y: Tensor, **kwargs: Any
    ) -> BatchedMultiOutputGPyTorchModel:
        """Conditions on additional observations for a Fully Bayesian model (either
        identical across models or unique per-model).

        Args:
            X: A `batch_shape x num_samples x d`-dim Tensor, where `d` is
                the dimension of the feature space and `batch_shape` is the number of
                sampled models.
            Y: A `batch_shape x num_samples x 1`-dim Tensor, where `d` is
                the dimension of the feature space and `batch_shape` is the number of
                sampled models.

        Returns:
            BatchedMultiOutputGPyTorchModel: A fully bayesian model conditioned on
              given observations. The returned model has `batch_shape` copies of the
              training data in case of identical observations (and `batch_shape`
              training datasets otherwise).
        """
        if X.ndim == 2 and Y.ndim == 2:
            # To avoid an error in GPyTorch when inferring the batch dimension, we add
            # the explicit batch shape here. The result is that the conditioned model
            # will have 'batch_shape' copies of the training data.
            X = X.repeat(self.batch_shape + (1, 1))
            Y = Y.repeat(self.batch_shape + (1, 1))

        elif X.ndim < Y.ndim:
            # We need to duplicate the training data to enable correct batch
            # size inference in gpytorch.
            X = X.repeat(*(Y.shape[:-2] + (1, 1)))

        return super().condition_on_observations(X, Y, **kwargs)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Abstract base module for all botorch posteriors.
"""

from __future__ import annotations

from typing import Callable, Optional, Union

import torch
from botorch.posteriors.posterior import Posterior
from botorch.sampling.get_sampler import _get_sampler_mvn, GetSampler
from botorch.sampling.normal import NormalMCSampler
from gpytorch.distributions.multivariate_normal import MultivariateNormal
from torch import Tensor


class BoundedRiemannPosterior(Posterior):
    batch_range = (0, -1)

    """
    A single variate bounded Riemann posterior.
    """

    def __init__(self, borders, probabilities):
        r"""Bounded Riemann Posterior.

        A posterior distribution represented by a piecewise constant probability density
        function over a bounded domain. The domain is divided into buckets defined by
        borders, with each bucket having an associated probability.

        Args:
            borders: A tensor of shape `(num_buckets + 1,)` defining the boundaries of
                the buckets. Must be monotonically increasing.
            probabilities: A tensor of shape `(b?, q?, num_buckets)` defining the
                probability mass in each bucket. Must sum to 1 in the last dim.
        """

        assert torch.allclose(
            probabilities.sum(-1),
            torch.tensor(1.0, device=probabilities.device, dtype=probabilities.dtype),
            atol=1e-3,
        ), f"Probabilities must sum to 1, but sum to {probabilities.sum()}."

        self.borders = borders
        self.probabilities = probabilities
        self.cumprobs = torch.cumsum(self.probabilities, -1)

    def integrate(self, ag_integrate_fn: Callable[[Tensor, Tensor], Tensor]) -> Tensor:
        r"""Integrate over the posterior, by calculating
            $$\int_{min}^{max} ag(y) posterior(y) dy$$.

        Args:
            ag_integral_fn: A vectorized function that integrates ag from lower_bound
                to upper_bound, that is, $$\int_{lower_bound}^{upper_bound} ag(y) dy$$.

        Returns:
            Tensor: The integral of the posterior.
        """
        all_lower = self.borders[:-1]
        all_upper = self.borders[1:]
        bucket_results = ag_integrate_fn(all_lower, all_upper)
        return (bucket_results * (self.probabilities / (all_upper - all_lower))).sum(-1)

    def rsample(
        self,
        sample_shape: Optional[torch.Size] = None,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        sample_shape = sample_shape if sample_shape is not None else torch.Size([1])
        base_samples = torch.randn(
            sample_shape + self.probabilities.shape[:-1],
            device=self.probabilities.device,
        )
        return self.rsample_from_base_samples(
            sample_shape=sample_shape, base_samples=base_samples
        )

    def rsample_from_base_samples(
        self,
        sample_shape: torch.Size,
        base_samples: Tensor,
    ) -> Tensor:
        """
        base_samples are N(0, I) samples, as this posterior is registered
        with the IIDNormalSampler below. Alternatively it could be registered
        with a uniform sampler in which case the transformation to uniform RVs
        could be avoided. Shape of base_samples is (nsamp, b?, q).
        """
        if base_samples.shape[: len(sample_shape)] != sample_shape:
            raise ValueError(
                "`sample_shape` disagrees with shape of `base_samples`. "
                f"Got {sample_shape=} and {base_samples.shape=}."
            )
        # convert base samples from N(O, I) to Uniform.
        U = torch.distributions.Normal(0, 1).cdf(base_samples)
        # Convert U to Riemann samples.
        Z = self.icdf(U)  # (nsamp, b?, q, 1)
        return Z

    @property
    def base_sample_shape(self) -> torch.Size:
        r"""The shape of the base samples required to draw from the posterior."""
        return self.probabilities.shape[:-1]

    @property
    def device(self) -> torch.device:
        r"""The torch device of the distribution."""
        return self.borders.device

    @property
    def dtype(self) -> torch.dtype:
        r"""The torch dtype of the distribution."""
        return self.borders.dtype

    @property
    def mean(self):
        r"""The mean of the posterior distribution."""
        bucket_widths = self.borders[1:] - self.borders[:-1]
        bucket_means = self.borders[:-1] + bucket_widths / 2
        return (self.probabilities @ bucket_means).unsqueeze(-1)

    @property
    def mean_of_square(self) -> torch.Tensor:
        """Computes E[x^2]."""
        left_borders = self.borders[:-1]
        right_borders = self.borders[1:]
        bucket_mean_of_square = (
            left_borders.square()
            + right_borders.square()
            + left_borders * right_borders
        ) / 3.0
        return (self.probabilities @ bucket_mean_of_square).unsqueeze(-1)

    @property
    def variance(self) -> torch.Tensor:
        """Computes the variance via Var[x] = E[x^2] - E[x]^2."""
        return self.mean_of_square - self.mean.square()

    def confidence_region(
        self, confidence_level: float = 0.95
    ) -> tuple[Tensor, Tensor]:
        r"""
        Compute the lower and upper bounds of the confidence region.
        Args:
            confidence_level: The probability between the bounds
                of the confidence interval/region.
                Use .954 for 2 sigma of a normal distribution.
        """
        side_probs = (1.0 - confidence_level) / 2
        lower = self.icdf(side_probs).squeeze()
        upper = self.icdf(1.0 - side_probs).squeeze()
        return lower, upper

    def icdf(
        self,
        value: Union[float, Tensor],
    ) -> Tensor:
        r"""Inverse cdf (with gradients).
        Use value to get the index of the bucket that contains the value
        and then interpolate between the left and right borders of the bucket

        Args:
            value: The value at which to evaluate the inverse CDF.
                Either a float, or a tensor with shape is (b', b?, q), where
                probabilities has shape (b?, q, num_buckets).

        Returns:
            The inverse CDF of the posterior at the given value(s).
            The shape of the return is (b', b?, q, 1), with a trailing
            dimension.
        """
        if not torch.is_tensor(value):
            # Promote to a (b', b?, q) tensor
            value = torch.tensor(value, device=self.device, dtype=self.dtype)
            value = value.expand(*self.probabilities.shape[:-1]).unsqueeze(0)
        value = value.movedim(0, -1)  # (b?, q, b')

        index = torch.searchsorted(self.cumprobs, value)  # (b?, q, b')

        left_border = self.borders[index]  # (b?, q, b')
        right_border = self.borders[index + 1]

        bucket_width = right_border - left_border
        right_cum_probs = torch.gather(self.cumprobs, -1, index)
        prob_width = torch.gather(self.probabilities, -1, index)

        bucket_proportion_remaining = (right_cum_probs - value) / prob_width
        result = (
            right_border - bucket_proportion_remaining * bucket_width
        )  # (b?, q, b')

        # reshape back to (b', b?, q, 1)
        result = result.movedim(-1, 0).unsqueeze(-1)
        return result


class MultivariateRiemannPosterior(BoundedRiemannPosterior):
    def __init__(self, borders, probabilities, correlation_matrix) -> None:
        """
        A multi-variate bounded Riemann posterior using a Gaussian copula.

        Uses BoundedRiemannPosterior for marginal distributions, and then MVN
        correlation structure via the Gaussian copula.

        Args:
            borders: A tensor of shape `(num_buckets + 1,)` defining the boundaries of
                the buckets. Must be monotonically increasing.
            probabilities: A tensor of shape `(b?, q, num_buckets)` defining the
                probability mass in each bucket. Must sum to 1 in the last dim.
            correlation_matrix: The Guassian correlation matrix, (b?, q, q).
        """
        super().__init__(borders=borders, probabilities=probabilities)
        self.correlation_matrix = correlation_matrix

    def rsample_from_base_samples(
        self, sample_shape: torch.Size, base_samples: Tensor
    ) -> Tensor:
        """
        Sample from the posterior using base samples.

        base_samples are N(0, I) samples, as this posterior is registered
        with the IIDNormalSampler below. This is also necessary for the use
        of the Gaussian copula.

        Args:
            sample_shape: Shape of samples.
            base_samples: (nsamp, b?, q) standard normal samples.

        Returns: Samples from copula, shape (nsamp, b?, q, 1)
        """
        # Construct MVN
        mvn = MultivariateNormal(
            mean=torch.zeros_like(self.mean.squeeze(-1)),
            covariance_matrix=self.correlation_matrix,
        )
        # Draw samples
        samples = mvn.rsample(
            sample_shape=sample_shape, base_samples=base_samples
        ).squeeze(-1)  # (nsamp, b?, q)
        # Convert N(0, 1) marginals to Uniform samples.
        U = torch.distributions.Normal(0, 1).cdf(samples)  # (nsamp, b?, q)
        Z = self.icdf(U)  # (nsamp, b?, q, 1)
        return Z


@GetSampler.register(BoundedRiemannPosterior)
def _get_sampler_riemann(
    posterior: BoundedRiemannPosterior,
    sample_shape: torch.Size,
    *,
    seed: int | None = None,
) -> NormalMCSampler:
    return _get_sampler_mvn(
        posterior=posterior,
        sample_shape=sample_shape,
        seed=seed,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
This file contains a readily usable implementation of the robust Gaussian process
model of [Ament2024pursuit]_, leveraging the Relevance Pursuit algorithm.

In particular, this file contains a `RobustRelevancePursuitMixin` class, and a concrete
implementation of a `SingleTaskGP` model, `RobustRelevancePursuitSingleTaskGP`, which
has the same API as a standard `SingleTaskGP` model, but automatically instantiates the
robust likelihood `SparseOutlierGaussianLikelihood` and dispatches the relevance pursuit
algorithm during model fitting via `fit_gpytorch_mll`.

Even though a standard `SingleTaskGP` model is expressive enough to implement the robust
model by changing the likelihood, its optimization is more complex. So the main reason
for the `RobustRelevancePursuitMixin` class is to hide this complexity by using multiple
dispatch of `fit_gpytorch_mll`, which needs to do two distinct operations in the context
of the robust model:

(1) It needs to toggle the relevance pursuit discrete optimization algorithm that
    changes the support, and as a sub-task,
(2) it needs to still carry out the numerical optimization of the hyper-parameters given
    a fixed support, but still with a `SparseOutlierGaussianLikelihood`. Since the types
    of the marginal likelihood (`MarginalLogLikelihood`) and the likelihood
    (`SparseOutlierGaussianLikelihood`) are the same in both calls, the only way we can
    leverage the multiple dispatch mechanism is the model type.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from typing import Any, Callable, Mapping, Optional, Sequence

import torch
from botorch.exceptions.errors import UnsupportedError
from botorch.fit import FitGPyTorchMLL
from botorch.models import SingleTaskGP
from botorch.models.likelihoods.sparse_outlier_noise import (
    SparseOutlierGaussianLikelihood,
    SparseOutlierNoise,
)
from botorch.models.model import Model
from botorch.models.relevance_pursuit import (
    backward_relevance_pursuit,
    get_posterior_over_support,
)
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.likelihoods import (
    FixedNoiseGaussianLikelihood,
    GaussianLikelihood,
    Likelihood,
)
from gpytorch.means.mean import Mean
from gpytorch.mlls._approximate_mll import _ApproximateMarginalLogLikelihood
from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
from gpytorch.module import Module
from torch import Tensor

# default fractions of outliers to consider during relevance pursuit
FRACTIONS_OF_OUTLIERS = [
    0.0,
    0.05,
    0.1,
    0.15,
    0.2,
    0.3,
    0.4,
    0.5,
    0.75,
    1.0,
]


class RobustRelevancePursuitMixin(ABC):
    """A Mixin class for robust relevance pursuit models, which wraps a base likelihood
    with a `SparseOutlierGaussianLikelihood` to detect outliers, and calls the
    relevance pursuit algorithm during model fitting via `fit_gpytorch_mll`.

    This is distinct from the `RelevancePursuitMixin` class, which is a Mixin class to
    equip a specific module (the likelihood, in the case of the robust model) with the
    relevance pursuit algorithms.
    """

    def __init__(
        self,
        base_likelihood: GaussianLikelihood | FixedNoiseGaussianLikelihood,
        dim: int,
        prior_mean_of_support: float | None = None,
        convex_parameterization: bool = True,
        cache_model_trace: bool = False,
    ) -> None:
        """Initializes a robust relevance pursuit model, which wraps a base likelihood
        with a `SparseOutlierGaussianLikelihood` to detect outliers, and calls the
        relevance pursuit algorithm during model fitting via `fit_gpytorch_mll`.

        For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

        Args:
            base_likelihood: The base likelihood that will be wrapped by a
                `SparseOutlierGaussianLikelihood` to detect outliers.
            dim: The number of training data points, i.e. the maximum dimensionality
                of the support set of the likelihood.
            prior_mean_of_support: The mean value for the default exponential prior
                distribution over the support size.
            convex_parameterization: If True, use a convex parameterization of the
                sparse noise model. See `SparseOutlierGaussianLikelihood` for details.
            cache_model_trace: If True, cache the model trace during relevance pursuit.
        """
        self.likelihood = SparseOutlierGaussianLikelihood(
            base_noise=base_likelihood.noise_covar,
            dim=dim,
            convex_parameterization=convex_parameterization,
        )
        self.bmc_support_sizes: Tensor | None = None
        self.bmc_probabilities: Tensor | None = None
        self.cache_model_trace = cache_model_trace
        self.model_trace: list[SingleTaskGP] | None = None
        self.prior_mean_of_support: float = (
            int(0.2 * dim) if prior_mean_of_support is None else prior_mean_of_support
        )

    @abstractmethod
    def to_standard_model(self) -> Model:
        """Converts this `RobustRelevancePursuitMixin` to an equivalent standard model
        with the same robust likelihood and hyper-parameters. This leaves the model
        structure and predictions unchanged, but leads `fit_gpytorch_mll`'s dispatch to
        *numerically* optimize the hyper-parameters of the model with a fixed support
        set, as opposed to dispatching to the discrete optimization via the relevance
        pursuit algorithm.

        Returns:
            A standard model.
        """

    def load_standard_model(self, standard_model: Model) -> RobustRelevancePursuitMixin:
        """Loads the state dict of a model into the `RobustRelevancePursuitMixin`.

        Args:
            standard_model: A standard model with the same parameter structure and
                likelihood as the `RobustRelevancePursuitMixin` model.

        Returns:
            The `RobustRelevancePursuitMixin` with the standard model's state dict.
        """
        # need special case for the likelihood because raw_rho's shape changes
        # throughout the optimization
        self.likelihood = standard_model.likelihood
        # overwrite state_dict in place
        self.load_state_dict(standard_model.state_dict())
        return self


class RobustRelevancePursuitSingleTaskGP(SingleTaskGP, RobustRelevancePursuitMixin):
    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Tensor | None = None,
        likelihood: Likelihood | None = None,
        covar_module: Module | None = None,
        mean_module: Mean | None = None,
        outcome_transform: OutcomeTransform | _DefaultType | None = DEFAULT,
        input_transform: InputTransform | None = None,
        convex_parameterization: bool = True,
        prior_mean_of_support: float | None = None,
        cache_model_trace: bool = False,
    ) -> None:
        r"""A robust single-task GP model that toggles the relevance pursuit algorithm
            during model fitting via `fit_gpytorch_mll`.

        For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            train_Yvar: An optional `batch_shape x n x m` tensor of observed
                measurement noise.
            likelihood: A base likelihood that will be wrapped by a
                `SparseOutlierGaussianLikelihood` to detect outliers. If omitted,
                use a standard `GaussianLikelihood` with inferred noise level if
                `train_Yvar` is None, and a `FixedNoiseGaussianLikelihood` with the
                given noise observations if `train_Yvar` is not None.
            covar_module: The module computing the covariance (Kernel) matrix.
                If omitted, uses an `RBFKernel`.
            mean_module: The mean function to be used. If omitted, use a
                `ConstantMean`.
            outcome_transform: An outcome transform that is applied to the
                training data during instantiation and to the posterior during
                inference (that is, the `Posterior` obtained by calling
                `.posterior` on the model will be on the original scale). We use a
                `Standardize` transform if no `outcome_transform` is specified.
                Pass down `None` to use no outcome transform.
            input_transform: An input transform that is applied in the model's
                forward pass.
            convex_parameterization: If True, use a convex parameterization of the
                sparse noise model. See `SparseOutlierGaussianLikelihood` for details.
            prior_mean_of_support: The mean value for the default exponential prior
                distribution over the support size.
            cache_model_trace: If True, cache the model trace during relevance pursuit.

        Example:
            >>> m = RobustRelevancePursuitSingleTaskGP(train_X=X, train_Y=Y)
            >>> mll = ExactMarginalLogLikelihood(model=m, likelihood=m.likelihood)
            >>> mll = fit_gpytorch_mll(mll)
        """
        self._original_X = train_X
        self._original_Y = train_Y
        super().__init__(
            train_X=train_X,
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            likelihood=likelihood,
            covar_module=covar_module,
            mean_module=mean_module,
            outcome_transform=outcome_transform,
            input_transform=input_transform,
        )
        # After canonical GP is instantiated, modify the likelihood
        RobustRelevancePursuitMixin.__init__(
            self,
            base_likelihood=self.likelihood,
            dim=train_X.shape[-2],
            prior_mean_of_support=prior_mean_of_support,
            convex_parameterization=convex_parameterization,
            cache_model_trace=cache_model_trace,
        )

    def to_standard_model(self) -> Model:
        """Returns a standard SingleTaskGP with the same parameters as this model.
        This is used to avoid recursion through the fit_gpytorch_mll dispatch."""
        # don't need to put model into training mode to access the untransformed inputs,
        # since we cached the original train_inputs
        is_training = self.training
        model = SingleTaskGP(
            train_X=self._original_X,
            train_Y=self._original_Y,
            train_Yvar=None,  # not needed because likelihood is already instantiated
            likelihood=self.likelihood,
            covar_module=self.covar_module,
            mean_module=self.mean_module,
            outcome_transform=getattr(self, "outcome_transform", None),
            input_transform=getattr(self, "input_transform", None),
        )
        if not is_training:
            model.eval()
        return model


@FitGPyTorchMLL.register(
    MarginalLogLikelihood,
    SparseOutlierGaussianLikelihood,
    RobustRelevancePursuitMixin,
)
def _fit_rrp(
    mll: MarginalLogLikelihood,
    _: type[SparseOutlierGaussianLikelihood],
    __: type[RobustRelevancePursuitMixin],
    *,
    numbers_of_outliers: list[int] | None = None,
    fractions_of_outliers: list[float] | None = None,
    timeout_sec: float | None = None,
    relevance_pursuit_optimizer: Callable = backward_relevance_pursuit,
    reset_parameters: bool = True,
    reset_dense_parameters: bool = False,
    # fit_gpytorch_mll kwargs
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]] | None = None,
    optimizer: Callable | None = None,
    closure_kwargs: dict[str, Any] | None = None,
    optimizer_kwargs: Optional[Mapping[str, Any]] = None,
) -> MarginalLogLikelihood:
    """Fits a RobustRelevancePursuitGP model using the given marginal likelihood.

    For details, see [Ament2024pursuit]_ or https://arxiv.org/abs/2410.24222.

    Args:
        mll: The marginal likelihood to fit.
        _: A likelihood, only directly used for dispatching.
        _: A model, only directly used for dispatching.
        numbers_of_outliers: An optional list of numbers of outliers to consider during
            relevance pursuit. By default, the algorithm falls back to a default list
            of fractions of outliers, see below.
        fractions_of_outliers: An optional list of fractions of outliers to consider if
            numbers_of_outliers is None. By default, the algorithm uses
            `[0, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.75, 1.0]`.
        relevance_pursuit_optimizer: The relevance pursuit optimizer to use. By default,
            uses `backward_relevance_pursuit`, which is generally the most powerful
            algorithm for challenging problems with a wide range of outliers. The
            `forward_relevance_pursuit` algorithm can be efficient when the number of
            outliers is relatively small.
        reset_parameters: If True, we will reset the sparse parameters of the model
            after each iteration of the relevance pursuit algorithm.
        reset_dense_parameters: If True, we will reset the dense parameters of the model
            after each iteration of the relevance pursuit algorithm.
        closure: A closure to use to compute the loss and the gradients, see docstring
            of `fit_gpytorch_mll` for details.
        optimizer: The numerical optimizer, see docstring of `fit_gpytorch_mll`.
        closure_kwargs: Additional arguments to pass to the `closure` function.
        optimizer_kwargs: Additional arguments to pass to `fit_gpytorch_mll`.

    Returns:
        The fitted marginal likelihood.
    """
    sparse_module = SparseOutlierNoise._from_model(mll.model)
    n = sparse_module.dim  # equal to the number of training data points

    if numbers_of_outliers is None:
        if fractions_of_outliers is None:
            fractions_of_outliers = FRACTIONS_OF_OUTLIERS

        # list from which BMC chooses
        numbers_of_outliers = [int(p * n) for p in fractions_of_outliers]

    optimizer_kwargs_: dict[str, Any] = (
        {} if optimizer_kwargs is None else dict(optimizer_kwargs)
    )
    if timeout_sec is not None:
        optimizer_kwargs_["timeout_sec"] = timeout_sec / len(numbers_of_outliers)

    # Need to convert model to avoid recursion through fit_gpytorch_mll dispatch, since
    # relevance pursuit expects to call the base fit_gpytorch_mll.
    original_model = mll.model  # Robust Relevance Pursuit Model
    mll.model = original_model.to_standard_model()
    sparse_module = SparseOutlierNoise._from_model(mll.model)
    sparse_module, model_trace = relevance_pursuit_optimizer(
        sparse_module=sparse_module,
        mll=mll,
        sparsity_levels=numbers_of_outliers,
        reset_parameters=reset_parameters,
        reset_dense_parameters=reset_dense_parameters,
        record_model_trace=True,
        # These are the args of the canonical mll fit routine
        closure=closure,
        optimizer=optimizer,
        closure_kwargs=closure_kwargs,
        optimizer_kwargs=optimizer_kwargs_,
    )

    # Bayesian model comparison
    bmc_support_sizes, bmc_probabilities = get_posterior_over_support(
        SparseOutlierNoise,
        model_trace,
        prior_mean_of_support=original_model.prior_mean_of_support,
    )
    map_index = torch.argmax(bmc_probabilities)
    map_model = model_trace[map_index]  # choosing model with highest BMC score
    # overwrite mll.model with chosen model
    mll.model = original_model  # first restore original model pointer
    mll.model.load_standard_model(map_model)
    # Store the bmc results
    mll.model.bmc_support_sizes = bmc_support_sizes
    mll.model.bmc_probabilities = bmc_probabilities
    if mll.model.cache_model_trace:
        mll.model.model_trace = model_trace
    return mll


@FitGPyTorchMLL.register(
    _ApproximateMarginalLogLikelihood,
    SparseOutlierGaussianLikelihood,
    RobustRelevancePursuitMixin,
)
def _fit_rrp_approximate_mll(
    mll: _ApproximateMarginalLogLikelihood,
    _: type[SparseOutlierGaussianLikelihood],
    __: type[RobustRelevancePursuitMixin],
    **kwargs: Any,
) -> None:
    raise UnsupportedError(
        "Relevance Pursuit does not yet support approximate inference. "
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Abstract base module for decoupled acquisition functions."""

from __future__ import annotations

import warnings
from abc import ABC

import torch
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.exceptions import BotorchWarning
from botorch.exceptions.errors import BotorchTensorDimensionError
from botorch.logging import shape_to_str

from botorch.models.model import ModelList
from torch import Tensor


class DecoupledAcquisitionFunction(AcquisitionFunction, ABC):
    """
    Abstract base class for decoupled acquisition functions.
    A decoupled acquisition function where one may intend to
    evaluate a design on only a subset of the outcomes.
    Typically this would be handled by fantasizing, where one
    would fantasize as to what the partial observation would
    be if one were to evaluate a design on the subset of
    outcomes (e.g. you only fantasize at those outcomes). The
    `X_evaluation_mask` specifies which outcomes should be
    evaluated for each design.  `X_evaluation_mask` is `q x m`,
    where there are q design points in the batch and m outcomes.
    In the asynchronous case, where there are n' pending points,
    we need to track which outcomes each pending point should be
    evaluated on. In this case, we concatenate
    `X_pending_evaluation_mask` with `X_evaluation_mask` to obtain
    the full evaluation_mask.


    This abstract class handles generating and updating an evaluation mask,
    which is a boolean tensor indicating which outcomes a given design is
    being evaluated on. The evaluation mask has shape `(n' + q) x m`, where
    n' is the number of pending points and the q represents the new
    candidates to be generated.

    If `X(_pending)_evaluation_mas`k is None, it is assumed that `X(_pending)`
    will be evaluated on all outcomes.
    """

    def __init__(
        self, model: ModelList, X_evaluation_mask: Tensor | None = None, **kwargs
    ) -> None:
        r"""Initialize.

        Args:
            model: A model
            X_evaluation_mask: A `q x m`-dim boolean tensor
                indicating which outcomes the decoupled acquisition
                function should generate new candidates for.
        """
        if not isinstance(model, ModelList):
            raise ValueError(f"{self.__class__.__name__} requires using a ModelList.")
        super().__init__(model=model, **kwargs)
        self.num_outputs = model.num_outputs
        self.X_evaluation_mask = X_evaluation_mask
        self.X_pending_evaluation_mask = None
        self.X_pending = None

    @property
    def X_evaluation_mask(self) -> Tensor | None:
        r"""Get the evaluation indices for the new candidate."""
        return self._X_evaluation_mask

    @X_evaluation_mask.setter
    def X_evaluation_mask(self, X_evaluation_mask: Tensor | None = None) -> None:
        r"""Set the evaluation indices for the new candidate."""
        if X_evaluation_mask is not None:
            # TODO: Add batch support
            if (
                X_evaluation_mask.ndim != 2
                or X_evaluation_mask.shape[-1] != self.num_outputs
            ):
                raise BotorchTensorDimensionError(
                    "Expected X_evaluation_mask to be `q x m`, but got shape"
                    f" {shape_to_str(X_evaluation_mask.shape)}."
                )
        self._X_evaluation_mask = X_evaluation_mask

    def set_X_pending(
        self,
        X_pending: Tensor | None = None,
        X_pending_evaluation_mask: Tensor | None = None,
    ) -> None:
        r"""Informs the AF about pending design points for different outcomes.

        Args:
            X_pending: A `n' x d` Tensor with `n'` `d`-dim design points that have
                been submitted for evaluation but have not yet been evaluated.
            X_pending_evaluation_mask: A `n' x m`-dim tensor of booleans indicating
                for which outputs the pending point is being evaluated on. If
                `X_pending_evaluation_mask` is `None`, it is assumed that
                `X_pending` will be evaluated on all outcomes.
        """
        if X_pending is not None:
            if X_pending.requires_grad:
                warnings.warn(
                    "Pending points require a gradient but the acquisition function"
                    " will not provide a gradient to these points.",
                    BotorchWarning,
                    stacklevel=2,
                )
            self.X_pending = X_pending.detach().clone()
            if X_pending_evaluation_mask is not None:
                if (
                    X_pending_evaluation_mask.ndim != 2
                    or X_pending_evaluation_mask.shape[0] != X_pending.shape[0]
                    or X_pending_evaluation_mask.shape[1] != self.num_outputs
                ):
                    raise BotorchTensorDimensionError(
                        f"Expected `X_pending_evaluation_mask` of shape "
                        f"`{X_pending.shape[0]} x {self.num_outputs}`, but "
                        f"got {shape_to_str(X_pending_evaluation_mask.shape)}."
                    )
                self.X_pending_evaluation_mask = X_pending_evaluation_mask
            elif self.X_evaluation_mask is not None:
                raise ValueError(
                    "If `self.X_evaluation_mask` is not None, then "
                    "`X_pending_evaluation_mask` must be provided."
                )

        else:
            self.X_pending = X_pending
            self.X_pending_evaluation_mask = X_pending_evaluation_mask

    def construct_evaluation_mask(self, X: Tensor) -> Tensor | None:
        r"""Construct the boolean evaluation mask for X and X_pending

        Args:
            X: A `batch_shape x n x d`-dim tensor of designs.

        Returns:
            A `n + n' x m`-dim tensor of booleans indicating
            which outputs should be evaluated.
        """
        if self.X_pending_evaluation_mask is not None:
            X_evaluation_mask = self.X_evaluation_mask
            if X_evaluation_mask is None:
                # evaluate all objectives for X
                X_evaluation_mask = torch.ones(
                    X.shape[-2], self.num_outputs, dtype=torch.bool, device=X.device
                )
            elif X_evaluation_mask.shape[0] != X.shape[-2]:
                raise BotorchTensorDimensionError(
                    "Expected the -2 dimension of X and X_evaluation_mask to match."
                )
            # construct mask for X
            return torch.cat(
                [X_evaluation_mask, self.X_pending_evaluation_mask], dim=-2
            )
        return self.X_evaluation_mask



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Risk Measures implemented as Monte-Carlo objectives, based on Bayesian
optimization of risk measures as introduced in [Cakmak2020risk]_. For a
broader discussion of Monte-Carlo methods for VaR and CVaR risk measures,
see also [Hong2014review]_.

.. [Cakmak2020risk]
    S. Cakmak, R. Astudillo, P. Frazier, and E. Zhou. Bayesian Optimization of
    Risk Measures. Advances in Neural Information Processing Systems 33, 2020.

.. [Hong2014review]
    L. J. Hong, Z. Hu, and G. Liu. Monte carlo methods for value-at-risk and
    conditional value-at-risk: a review. ACM Transactions on Modeling and
    Computer Simulation, 2014.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable
from math import ceil

import torch
from botorch.acquisition.multi_objective.objective import IdentityMCMultiOutputObjective
from botorch.acquisition.objective import IdentityMCObjective, MCAcquisitionObjective
from torch import Tensor


class RiskMeasureMCObjective(MCAcquisitionObjective, ABC):
    r"""Objective transforming the posterior samples to samples of a risk measure.

    The risk measure is calculated over joint q-batch samples from the posterior.
    If the q-batch includes samples corresponding to multiple inputs, it is assumed
    that first `n_w` samples correspond to first input, second `n_w` samples
    correspond to second input etc.

    The risk measures are commonly defined for minimization by considering the
    upper tail of the distribution, i.e., treating larger values as being undesirable.
    BoTorch by default assumes a maximization objective, so the default behavior here
    is to calculate the risk measures w.r.t. the lower tail of the distribution.
    This can be changed by passing a preprocessing function with
    `weights=torch.tensor([-1.0])`.
    """

    def __init__(
        self,
        n_w: int,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
    ) -> None:
        r"""Transform the posterior samples to samples of a risk measure.

        Args:
            n_w: The size of the `w_set` to calculate the risk measure over.
            preprocessing_function: A preprocessing function to apply to the samples
                before computing the risk measure. This can be used to scalarize
                multi-output samples before calculating the risk measure.
                For constrained optimization, this should also apply
                feasibility-weighting to samples. Given a `batch x m`-dim
                tensor of samples, this should return a `batch`-dim tensor.
        """
        super().__init__()
        self.n_w = n_w
        if preprocessing_function is None:
            if self._is_mo:
                preprocessing_function = IdentityMCMultiOutputObjective()
            else:
                preprocessing_function = IdentityMCObjective()
        self.preprocessing_function = preprocessing_function

    def _prepare_samples(self, samples: Tensor) -> Tensor:
        r"""Prepare samples for risk measure calculations by scalarizing and
        separating out the q-batch dimension.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.

        Returns:
            A `sample_shape x batch_shape x q x n_w`-dim tensor of prepared samples.
        """
        if samples.shape[-1] > 1 and isinstance(
            self.preprocessing_function, IdentityMCObjective
        ):
            raise RuntimeError(
                "Multi-output samples should be scalarized using a "
                "`preprocessing_function`."
            )
        samples = self.preprocessing_function(samples)
        return samples.view(*samples.shape[:-1], -1, self.n_w)

    @abstractmethod
    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the risk measure corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q`-dim tensor of risk measure samples.
        """
        pass  # pragma: no cover


class CVaR(RiskMeasureMCObjective):
    r"""The Conditional Value-at-Risk risk measure.

    The Conditional Value-at-Risk measures the expectation of the worst outcomes
    (small rewards or large losses) with a total probability of `1 - alpha`. It
    is commonly defined as the conditional expectation of the reward function,
    with the condition that the reward is smaller than the corresponding
    Value-at-Risk (also defined below).

    Note: Due to the use of a discrete `w_set` of samples, the VaR and CVaR
        calculated here are (possibly biased) Monte-Carlo approximations of
        the true risk measures.
    """

    def __init__(
        self,
        alpha: float,
        n_w: int,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
    ) -> None:
        r"""Transform the posterior samples to samples of a risk measure.

        Args:
            alpha: The risk level, float in `(0.0, 1.0]`.
            n_w: The size of the `w_set` to calculate the risk measure over.
            preprocessing_function: A preprocessing function to apply to the samples
                before computing the risk measure. This can be used to scalarize
                multi-output samples before calculating the risk measure.
                For constrained optimization, this should also apply
                feasibility-weighting to samples. Given a `batch x m`-dim
                tensor of samples, this should return a `batch`-dim tensor.
        """
        super().__init__(n_w=n_w, preprocessing_function=preprocessing_function)
        if not 0 < alpha <= 1:
            raise ValueError("alpha must be in (0.0, 1.0]")
        self.alpha = alpha
        self.alpha_idx = ceil(n_w * alpha) - 1

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the CVaR corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q`-dim tensor of CVaR samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return torch.topk(
            prepared_samples,
            k=prepared_samples.shape[-1] - self.alpha_idx,
            largest=False,
            dim=-1,
        ).values.mean(dim=-1)


class VaR(CVaR):
    r"""The Value-at-Risk risk measure.

    Value-at-Risk measures the smallest possible reward (or largest possible loss)
    after excluding the worst outcomes with a total probability of `1 - alpha`. It
    is commonly used in financial risk management, and it corresponds to the
    `1 - alpha` quantile of a given random variable.
    """

    def __init__(
        self,
        alpha: float,
        n_w: int,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
    ) -> None:
        r"""Transform the posterior samples to samples of a risk measure.

        Args:
            alpha: The risk level, float in `(0.0, 1.0]`.
            n_w: The size of the `w_set` to calculate the risk measure over.
            preprocessing_function: A preprocessing function to apply to the samples
                before computing the risk measure. This can be used to scalarize
                multi-output samples before calculating the risk measure.
                For constrained optimization, this should also apply
                feasibility-weighting to samples. Given a `batch x m`-dim
                tensor of samples, this should return a `batch`-dim tensor.
        """
        super().__init__(
            n_w=n_w,
            alpha=alpha,
            preprocessing_function=preprocessing_function,
        )
        self._q = 1 - self.alpha_idx / n_w

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the VaR corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q`-dim tensor of VaR samples.
        """
        prepared_samples = self._prepare_samples(samples)
        # this is equivalent to sorting along dim=-1 in descending order
        # and taking the values at index self.alpha_idx. E.g.
        # >>> sorted_res = prepared_samples.sort(dim=-1, descending=True)
        # >>> sorted_res.values[..., self.alpha_idx]
        # Using quantile is far more memory efficient since `torch.sort`
        # produces values and indices tensors with shape
        # `sample_shape x batch_shape x (q * n_w) x m`
        return torch.quantile(
            input=prepared_samples,
            q=self._q,
            dim=-1,
            keepdim=False,
            interpolation="lower",
        )


class WorstCase(RiskMeasureMCObjective):
    r"""The worst-case risk measure."""

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the worst-case measure corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q`-dim tensor of worst-case samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return prepared_samples.min(dim=-1).values


class Expectation(RiskMeasureMCObjective):
    r"""The expectation risk measure.

    For unconstrained problems, we recommend using the `ExpectationPosteriorTransform`
    instead. `ExpectationPosteriorTransform` directly transforms the posterior
    distribution over `q * n_w` to a posterior of `q` expectations, significantly
    reducing the cost of posterior sampling as a result.
    """

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the expectation corresponding to the given samples.
        This calculates the expectation / mean / average of each `n_w` samples
        across the q-batch dimension. If `self.weights` is given, the samples
        are scalarized across the output dimension before taking the expectation.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q`-dim tensor of expectation samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return prepared_samples.mean(dim=-1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from botorch.sampling.pathwise.features.generators import gen_kernel_features
from botorch.sampling.pathwise.features.maps import (
    FeatureMap,
    KernelEvaluationMap,
    KernelFeatureMap,
)

__all__ = [
    "FeatureMap",
    "gen_kernel_features",
    "KernelEvaluationMap",
    "KernelFeatureMap",
]
